# pylint: disable=arguments-differ,line-too-long,missing-docstring,missing-module-docstring
import math
import random

# code adapted from:
# https://github.com/kakaobrain/fast-autoaugment
# https://github.com/rpmcruz/autoaugment
class EfficientNetRandomCrop:
    # pylint: disable=misplaced-comparison-constant
    def __init__(self, imgsize, min_covered=0.1, aspect_ratio_range=(3./4, 4./3),
                 area_range=(0.1, 1.0), max_attempts=10):
        assert 0.0 < min_covered
        assert 0 < aspect_ratio_range[0] <= aspect_ratio_range[1]
        assert 0 < area_range[0] <= area_range[1]
        assert 1 <= max_attempts

        self.min_covered = min_covered
        self.aspect_ratio_range = aspect_ratio_range
        self.area_range = area_range
        self.max_attempts = max_attempts
        self._fallback = EfficientNetCenterCrop(imgsize)

    def __call__(self, img):
        # https://github.com/tensorflow/tensorflow/blob/9274bcebb31322370139467039034f8ff852b004/tensorflow/core/kernels/sample_distorted_bounding_box_op.cc#L111
        original_width, original_height = img.size
        min_area = self.area_range[0] * (original_width * original_height)
        max_area = self.area_range[1] * (original_width * original_height)

        for _ in range(self.max_attempts):
            aspect_ratio = random.uniform(*self.aspect_ratio_range)
            height = int(round(math.sqrt(min_area / aspect_ratio)))
            max_height = int(round(math.sqrt(max_area / aspect_ratio)))

            if max_height * aspect_ratio > original_width:
                max_height = (original_width + 0.5 - 1e-7) / aspect_ratio
                max_height = int(max_height)
                if max_height * aspect_ratio > original_width:
                    max_height -= 1

            if max_height > original_height:
                max_height = original_height

            if height >= max_height:
                height = max_height

            height = int(round(random.uniform(height, max_height)))
            width = int(round(height * aspect_ratio))
            area = width * height

            if area < min_area or area > max_area:
                continue
            if width > original_width or height > original_height:
                continue
            if area < self.min_covered * (original_width * original_height):
                continue
            if width == original_width and height == original_height:
                return self._fallback(img)      # https://github.com/tensorflow/tpu/blob/master/models/official/efficientnet/preprocessing.py#L102

            x = random.randint(0, original_width - width)
            y = random.randint(0, original_height - height)
            return img.crop((x, y, x + width, y + height))

        return self._fallback(img)


class EfficientNetCenterCrop:
    def __init__(self, imgsize):
        self.imgsize = imgsize

    def __call__(self, img):
        """Crop the given PIL Image and resize it to desired size.
        Args:
            img (PIL Image): Image to be cropped. (0,0) denotes the top left corner of the image.
            output_size (sequence or int): (height, width) of the crop box. If int,
                it is used for both directions
        Returns:
            PIL Image: Cropped image.
        """
        image_width, image_height = img.size
        image_short = min(image_width, image_height)

        crop_size = float(self.imgsize) / (self.imgsize + 32) * image_short

        crop_height, crop_width = crop_size, crop_size
        crop_top = int(round((image_height - crop_height) / 2.))
        crop_left = int(round((image_width - crop_width) / 2.))
        return img.crop((crop_left, crop_top, crop_left + crop_width, crop_top + crop_height))



"""Loss layers for keypoints that can be inserted to modules"""
import torch
import torch.nn as nn

__all__ = ['WeightedMSELoss', 'HMFocalLoss']

def _sigmoid(x):
    y = torch.clamp(x.sigmoid_(), min=1e-4, max=1-1e-4)
    return y

class WeightedMSELoss(nn.Module):
    """Weighted MSE loss layer"""
    def __init__(self):
        super().__init__()

    def forward(self, pred, gt, mask):
        assert pred.size() == gt.size()
        loss = ((pred - gt) **2) * mask
        loss = loss.mean()
        return loss

class HMFocalLoss(nn.Module):
    """Heatmap Focal Loss layer"""
    def __init__(self, alpha, beta):
        super(HMFocalLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta

    def forward(self, pred, gt):
        pos_inds = gt.eq(1).float()
        neg_inds = gt.lt(1).float()

        pred = _sigmoid(pred)
        neg_weights = torch.pow(1 - gt, self.beta)

        pos_loss = torch.log(pred) * torch.pow(1 - pred, self.alpha) * pos_inds
        neg_loss = torch.log(1 - pred) * torch.pow(pred, self.alpha) * neg_weights * neg_inds

        num_pos = pos_inds.float().sum()
        pos_loss = pos_loss.sum()
        neg_loss = neg_loss.sum()

        if num_pos == 0:
            return -neg_loss
        else:
            return -(pos_loss + neg_loss) / num_pos



"""Experimental image transformations."""
from __future__ import division
import random
import numpy as np
import mxnet as mx
from mxnet import nd

def random_color_distort(src, brightness_delta=32, contrast_low=0.5, contrast_high=1.5,
                         saturation_low=0.5, saturation_high=1.5, hue_delta=18):
    """Randomly distort image color space.
    Note that input image should in original range [0, 255].

    Parameters
    ----------
    src : mxnet.nd.NDArray
        Input image as HWC format.
    brightness_delta : int
        Maximum brightness delta. Defaults to 32.
    contrast_low : float
        Lowest contrast. Defaults to 0.5.
    contrast_high : float
        Highest contrast. Defaults to 1.5.
    saturation_low : float
        Lowest saturation. Defaults to 0.5.
    saturation_high : float
        Highest saturation. Defaults to 1.5.
    hue_delta : int
        Maximum hue delta. Defaults to 18.

    Returns
    -------
    mxnet.nd.NDArray
        Distorted image in HWC format.

    """
    def brightness(src, delta, p=0.5):
        """Brightness distortion."""
        if np.random.uniform(0, 1) > p:
            delta = np.random.uniform(-delta, delta)
            src += delta
            return src
        return src

    def contrast(src, low, high, p=0.5):
        """Contrast distortion"""
        if np.random.uniform(0, 1) > p:
            alpha = np.random.uniform(low, high)
            src *= alpha
            return src
        return src

    def saturation(src, low, high, p=0.5):
        """Saturation distortion."""
        if np.random.uniform(0, 1) > p:
            alpha = np.random.uniform(low, high)
            gray = src * nd.array([[[0.299, 0.587, 0.114]]], ctx=src.context)
            gray = mx.nd.sum(gray, axis=2, keepdims=True)
            gray *= (1.0 - alpha)
            src *= alpha
            src += gray
            return src
        return src

    def hue(src, delta, p=0.5):
        """Hue distortion"""
        if np.random.uniform(0, 1) > p:
            alpha = random.uniform(-delta, delta)
            u = np.cos(alpha * np.pi)
            w = np.sin(alpha * np.pi)
            bt = np.array([[1.0, 0.0, 0.0],
                           [0.0, u, -w],
                           [0.0, w, u]])
            tyiq = np.array([[0.299, 0.587, 0.114],
                             [0.596, -0.274, -0.321],
                             [0.211, -0.523, 0.311]])
            ityiq = np.array([[1.0, 0.956, 0.621],
                              [1.0, -0.272, -0.647],
                              [1.0, -1.107, 1.705]])
            t = np.dot(np.dot(ityiq, bt), tyiq).T
            src = nd.dot(src, nd.array(t, ctx=src.context))
            return src
        return src

    src = src.astype('float32')

    # brightness
    src = brightness(src, brightness_delta)

    # color jitter
    if np.random.randint(0, 2):
        src = contrast(src, contrast_low, contrast_high)
        src = saturation(src, saturation_low, saturation_high)
        src = hue(src, hue_delta)
    else:
        src = saturation(src, saturation_low, saturation_high)
        src = hue(src, hue_delta)
        src = contrast(src, contrast_low, contrast_high)
    return src

_data_rng = np.random.RandomState(None)

def np_random_color_distort(image, data_rng=None, eig_val=None,
                            eig_vec=None, var=0.4, alphastd=0.1):
    """Numpy version of random color jitter.

    Parameters
    ----------
    image : numpy.ndarray
        original image.
    data_rng : numpy.random.rng
        Numpy random number generator.
    eig_val : numpy.ndarray
        Eigen values.
    eig_vec : numpy.ndarray
        Eigen vectors.
    var : float
        Variance for the color jitters.
    alphastd : type
        Jitter for the brightness.

    Returns
    -------
    numpy.ndarray
        The jittered image

    """
    from ....utils.filesystem import try_import_cv2
    cv2 = try_import_cv2()
    if data_rng is None:
        data_rng = _data_rng
    if eig_val is None:
        eig_val = np.array([0.2141788, 0.01817699, 0.00341571],
                           dtype=np.float32)
    if eig_vec is None:
        eig_vec = np.array([[-0.58752847, -0.69563484, 0.41340352],
                            [-0.5832747, 0.00994535, -0.81221408],
                            [-0.56089297, 0.71832671, 0.41158938]], dtype=np.float32)
    def grayscale(image):
        return cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

    def lighting_(data_rng, image, alphastd, eigval, eigvec):
        alpha = data_rng.normal(scale=alphastd, size=(3, ))
        image += np.dot(eigvec, eigval * alpha)

    def blend_(alpha, image1, image2):
        image1 *= alpha
        image2 *= (1 - alpha)
        image1 += image2

    def saturation_(data_rng, image, gs, gs_mean, var):
        # pylint: disable=unused-argument
        alpha = 1. + data_rng.uniform(low=-var, high=var)
        blend_(alpha, image, gs[:, :, None])

    def brightness_(data_rng, image, gs, gs_mean, var):
        # pylint: disable=unused-argument
        alpha = 1. + data_rng.uniform(low=-var, high=var)
        image *= alpha

    def contrast_(data_rng, image, gs, gs_mean, var):
        # pylint: disable=unused-argument
        alpha = 1. + data_rng.uniform(low=-var, high=var)
        blend_(alpha, image, gs_mean)

    functions = [brightness_, contrast_, saturation_]
    random.shuffle(functions)

    gs = grayscale(image)
    gs_mean = gs.mean()
    for f in functions:
        f(data_rng, image, gs, gs_mean, var)
    lighting_(data_rng, image, alphastd, eig_val, eig_vec)
    return image



"""Random wrapper."""
from __future__ import absolute_import
import random as pyrandom
import numpy as np
try:
    import mxnet as mx
except ImportError:
    mx = None


def seed(a=None):
    """Seed the generator for python builtin random, numpy.random, mxnet.random.

    This method is to control random state for mxnet related random functions.

    Note that this function cannot guarantee 100 percent reproducibility due to
    hardware settings.

    Parameters
    ----------
    a : int or 1-d array_like, optional
        Initialize internal state of the random number generator.
        If `seed` is not None or an int or a long, then hash(seed) is used instead.
        Note that the hash values for some types are nondeterministic.

    """
    pyrandom.seed(a)
    np.random.seed(a)
    if mx is not None:
        mx.random.seed(a)



from typing import Union
import torch

class Beziers:
    """
    This structure stores a list of bezier curves as a Nx16 torch.Tensor.
    It will support some common methods about bezier shapes
    (`area`, `clip`, `nonempty`, etc),
    and also behaves like a Tensor
    (support indexing, `to(device)`, `.device`, and iteration over all boxes)

    Attributes:
        tensor (torch.Tensor): float matrix of Nx4. Each row is (x1, y1, x2, y2).
    """

    def __init__(self, tensor: torch.Tensor):
        """
        Args:
            tensor (Tensor[float]): a Nx4 matrix.  Each row is (x1, y1, x2, y2).
        """
        device = tensor.device if isinstance(tensor, torch.Tensor) else torch.device("cpu")
        tensor = torch.as_tensor(tensor, dtype=torch.float32, device=device)
        if tensor.numel() == 0:
            # Use reshape, so we don't end up creating a new tensor that does not depend on
            # the inputs (and consequently confuses jit)
            tensor = tensor.reshape((0, 16)).to(dtype=torch.float32, device=device)
        assert tensor.dim() == 2 and tensor.size(-1) == 16, tensor.size()

        self.tensor = tensor

    def to(self, device: str) -> "Beziers":
        return Beziers(self.tensor.to(device))

    def __getitem__(self, item: Union[int, slice, torch.BoolTensor]) -> "Beziers":
        """
        Returns:
            Beziers: Create a new :class:`Beziers` by indexing.
        """
        if isinstance(item, int):
            return Beziers(self.tensor[item].view(1, -1))
        b = self.tensor[item]
        assert b.dim() == 2, "Indexing on Boxes with {} failed to return a matrix!".format(item)
        return Beziers(b)



# pylint: disable=wildcard-import
from timm.optim.optim_factory import *

_TIMM_FC_NAMES = ('fc', 'head', 'last_linear', 'classifier')

def optimizer_kwargs(cfg):
    kwargs = dict(
        opt=cfg.optimizer.opt,
        lr=cfg.train.lr,
        weight_decay=cfg.optimizer.weight_decay,
        momentum=cfg.optimizer.momentum)
    if cfg.optimizer.opt_eps is not None:
        kwargs['eps'] = cfg.optimizer.opt_eps
    if cfg.optimizer.opt_betas is not None:
        kwargs['betas'] = cfg.optimizer.opt_betas
    return kwargs

def add_weight_decay_filter_fc(model, learning_rate=0.001, weight_decay=1e-5, skip_list=(),
                               fc_names=_TIMM_FC_NAMES, feature_lr_mult=0.01, fc_lr_mult=1):
    if feature_lr_mult == 1 and fc_lr_mult == 1:
        return add_weight_decay(model, weight_decay, skip_list)
    decay_feat = []
    no_decay_feat = []
    decay_fc = []
    no_decay_fc = []
    for name, param in model.named_parameters():
        is_fc = False
        if any([name.startswith(pattern) for pattern in fc_names]):
            is_fc = True
        if not param.requires_grad:
            continue  # frozen weights
        if len(param.shape) == 1 or name.endswith(".bias") or name in skip_list:
            if is_fc:
                no_decay_fc.append(param)
            else:
                no_decay_feat.append(param)
        else:
            if is_fc:
                decay_fc.append(param)
            else:
                decay_feat.append(param)
    return [
        {'params': no_decay_feat, 'weight_decay': 0., 'lr': feature_lr_mult * learning_rate},
        {'params': no_decay_fc, 'weight_decay': 0., 'lr': fc_lr_mult * learning_rate},
        {'params': decay_feat, 'weight_decay': weight_decay, 'lr': feature_lr_mult * learning_rate},
        {'params': decay_fc, 'weight_decay': weight_decay, 'lr': fc_lr_mult * learning_rate}]

def filter_fc_layer(model, learning_rate=0.001, fc_names=_TIMM_FC_NAMES, feature_lr_mult=0.01, fc_lr_mult=1):
    """Filter linear projection layers from the network"""
    if feature_lr_mult == 1 and fc_lr_mult == 1:
        return model.parameters()
    feat = []
    fc = []
    for name, param in model.named_parameters():
        is_fc = False
        if any([name.startswith(pattern) for pattern in fc_names]):
            is_fc = True
        if not param.requires_grad:
            continue  # frozen weights
        if is_fc:
            fc.append(param)
        else:
            feat.append(param)

    return [
        {'params': feat, 'lr': feature_lr_mult * learning_rate},
        {'params': fc, 'lr': fc_lr_mult * learning_rate}]

def create_optimizer_v2a(
        model: nn.Module,
        optimizer_name: str = 'sgd',
        learning_rate: Optional[float] = None,
        weight_decay: float = 0.,
        momentum: float = 0.9,
        filter_bias_and_bn: bool = True,
        feature_lr_mult: float = 0.01,
        fc_lr_mult: float = 1,
        **kwargs):
    """ Create an optimizer.
    Note that this version is modifed based on
    https://github.com/rwightman/pytorch-image-models/blob/cd3dc4979f6ca16a09910b4a32b7a8f07cc31fda/timm/optim/optim_factory.py#L73
    It allows feature backbone and output linear layers to have different learning rate by having more groups
    TODO currently the model is passed in and all parameters are selected for optimization.
    For more general use an interface that allows selection of parameters to optimize and lr groups, one of:
      * a filter fn interface that further breaks params into groups in a weight_decay compatible fashion
      * expose the parameters interface and leave it up to caller
    Args:
        model (nn.Module): model containing parameters to optimize
        optimizer_name: name of optimizer to create
        learning_rate: initial learning rate
        weight_decay: weight decay to apply in optimizer
        momentum:  momentum for momentum based optimizers (others may use betas via kwargs)
        filter_bias_and_bn:  filter out bias, bn and other 1d params from weight decay
        **kwargs: extra optimizer specific kwargs to pass through
    Returns:
        Optimizer
    """
    opt_lower = optimizer_name.lower()
    if weight_decay and filter_bias_and_bn:
        skip = {}
        if hasattr(model, 'no_weight_decay'):
            skip = model.no_weight_decay()
        parameters = add_weight_decay_filter_fc(model, learning_rate, weight_decay, skip,
                                                _TIMM_FC_NAMES, feature_lr_mult, fc_lr_mult)
        weight_decay = 0.
    else:
        parameters = filter_fc_layer(model, learning_rate, _TIMM_FC_NAMES, feature_lr_mult, fc_lr_mult)
    if 'fused' in opt_lower:
        assert has_apex and torch.cuda.is_available(), 'APEX and CUDA required for fused optimizers'

    opt_args = dict(lr=learning_rate, weight_decay=weight_decay, **kwargs)
    opt_split = opt_lower.split('_')
    opt_lower = opt_split[-1]
    if opt_lower in ('sgd', 'nesterov'):
        opt_args.pop('eps', None)
        optimizer = optim.SGD(parameters, momentum=momentum, nesterov=True, **opt_args)
    elif opt_lower == 'momentum':
        opt_args.pop('eps', None)
        optimizer = optim.SGD(parameters, momentum=momentum, nesterov=False, **opt_args)
    elif opt_lower == 'adam':
        optimizer = optim.Adam(parameters, **opt_args)
    elif opt_lower == 'adabelief':
        optimizer = AdaBelief(parameters, rectify=False, **opt_args)
    elif opt_lower == 'adamw':
        optimizer = optim.AdamW(parameters, **opt_args)
    elif opt_lower == 'nadam':
        optimizer = Nadam(parameters, **opt_args)
    elif opt_lower == 'radam':
        optimizer = RAdam(parameters, **opt_args)
    elif opt_lower == 'adamp':
        optimizer = AdamP(parameters, wd_ratio=0.01, nesterov=True, **opt_args)
    elif opt_lower == 'sgdp':
        optimizer = SGDP(parameters, momentum=momentum, nesterov=True, **opt_args)
    elif opt_lower == 'adadelta':
        optimizer = optim.Adadelta(parameters, **opt_args)
    elif opt_lower == 'adafactor':
        if not learning_rate:
            opt_args['lr'] = None
        optimizer = Adafactor(parameters, **opt_args)
    elif opt_lower == 'adahessian':
        optimizer = Adahessian(parameters, **opt_args)
    elif opt_lower == 'rmsprop':
        optimizer = optim.RMSprop(parameters, alpha=0.9, momentum=momentum, **opt_args)
    elif opt_lower == 'rmsproptf':
        optimizer = RMSpropTF(parameters, alpha=0.9, momentum=momentum, **opt_args)
    elif opt_lower == 'novograd':
        optimizer = NovoGrad(parameters, **opt_args)
    elif opt_lower == 'nvnovograd':
        optimizer = NvNovoGrad(parameters, **opt_args)
    elif opt_lower == 'fusedsgd':
        opt_args.pop('eps', None)
        optimizer = FusedSGD(parameters, momentum=momentum, nesterov=True, **opt_args)
    elif opt_lower == 'fusedmomentum':
        opt_args.pop('eps', None)
        optimizer = FusedSGD(parameters, momentum=momentum, nesterov=False, **opt_args)
    elif opt_lower == 'fusedadam':
        optimizer = FusedAdam(parameters, adam_w_mode=False, **opt_args)
    elif opt_lower == 'fusedadamw':
        optimizer = FusedAdam(parameters, adam_w_mode=True, **opt_args)
    elif opt_lower == 'fusedlamb':
        optimizer = FusedLAMB(parameters, **opt_args)
    elif opt_lower == 'fusednovograd':
        opt_args.setdefault('betas', (0.95, 0.98))
        optimizer = FusedNovoGrad(parameters, **opt_args)
    else:
        assert False and "Invalid optimizer"
        raise ValueError

    if len(opt_split) > 1:
        if opt_split[0] == 'lookahead':
            optimizer = Lookahead(optimizer)

    return optimizer



"""Detection model utils"""
import torch
from torch import nn

from ...data.structures import Instances
from ...utils.utils import retry_if_cuda_oom

def build_top_module(cfg):
    """Build detection head/top module"""
    top_type = cfg.CONFIG.MODEL.DETECTION_TOP_MODULE.NAME
    if top_type == "conv":
        inp = cfg.CONFIG.MODEL.FPN.OUT_CHANNELS
        oup = cfg.CONFIG.MODEL.DETECTION_TOP_MODULE.DIM
        top_module = nn.Conv2d(
            inp, oup,
            kernel_size=3, stride=1, padding=1)
    else:
        top_module = None
    return top_module

def detector_postprocess(results, output_height, output_width, mask_threshold=0.5):
    """
    Resize the output instances.
    The input images are often resized when entering an object detector.
    As a result, we often need the outputs of the detector in a different
    resolution from its inputs.

    This function will resize the raw outputs of an R-CNN detector
    to produce outputs according to the desired output resolution.

    Args:
        results (Instances): the raw outputs from the detector.
            `results.image_size` contains the input image resolution the detector sees.
            This object might be modified in-place.
        output_height, output_width: the desired output resolution.

    Returns:
        Instances: the resized output from the model, based on the output resolution
    """

    # Converts integer tensors to float temporaries
    #   to ensure true division is performed when
    #   computing scale_x and scale_y.
    if isinstance(output_width, torch.Tensor):
        output_width_tmp = output_width.float()
    else:
        output_width_tmp = output_width

    if isinstance(output_height, torch.Tensor):
        output_height_tmp = output_height.float()
    else:
        output_height_tmp = output_height

    scale_x, scale_y = (
        output_width_tmp / results.image_size[1],
        output_height_tmp / results.image_size[0],
    )
    results = Instances((output_height, output_width), **results.get_fields())

    if results.has("pred_boxes"):
        output_boxes = results.pred_boxes
    elif results.has("proposal_boxes"):
        output_boxes = results.proposal_boxes

    output_boxes.scale(scale_x, scale_y)
    output_boxes.clip(results.image_size)

    results = results[output_boxes.nonempty()]

    if results.has("pred_masks"):
        results.pred_masks = retry_if_cuda_oom(paste_masks_in_image)(
            results.pred_masks[:, 0, :, :],  # N, 1, M, M
            results.pred_boxes,
            results.image_size,
            threshold=mask_threshold,
        )

    if results.has("pred_keypoints"):
        results.pred_keypoints[:, :, 0] *= scale_x
        results.pred_keypoints[:, :, 1] *= scale_y

    return results



"""Modified NMS ops as static/dynamic layers"""
import numpy as np
import torch
from torchvision.ops import boxes as box_ops
from torchvision.ops import nms  # BC-compat


__all__ = ['batched_nms', 'ml_nms']


def batched_nms(boxes: torch.Tensor, scores: torch.Tensor, idxs: torch.Tensor, iou_threshold: float):
    """
    Same as torchvision.ops.boxes.batched_nms, but safer.
    """
    assert boxes.shape[-1] == 4
    # TODO may need better strategy.
    # Investigate after having a fully-cuda NMS op.
    if len(boxes) < 40000:
        return box_ops.batched_nms(boxes, scores, idxs, iou_threshold)

    result_mask = scores.new_zeros(scores.size(), dtype=torch.bool)
    for iid in torch.unique(idxs).cpu().tolist():
        mask = (idxs == iid).nonzero().view(-1)
        keep = nms(boxes[mask], scores[mask], iou_threshold)
        result_mask[mask[keep]] = True
    keep = result_mask.nonzero().view(-1)
    keep = keep[scores[keep].argsort(descending=True)]
    return keep


def ml_nms(boxlist, nms_thresh, max_proposals=-1, fixed_size=False):
    """
    Performs non-maximum suppression on a boxlist, with scores specified
    in a boxlist field via score_field.

    Args:
        boxlist (data.structures.Boxes):
        nms_thresh (float): the score threshold for nms
        max_proposals (int): if > 0, then only the top max_proposals are kept
            after non-maximum suppression
        fixed_size (bool): force output to be static shape
    """
    if nms_thresh <= 0:
        return boxlist
    boxes = boxlist.pred_boxes.tensor
    scores = boxlist.scores
    labels = boxlist.pred_classes
    keep = batched_nms(boxes, scores, labels, nms_thresh)
    # keep = nms(boxes, scores, nms_thresh)
    if max_proposals > 0:
        keep = keep[: max_proposals]
    if fixed_size:
        boxlist._keep = keep
        # the following code block is one way to perform a static nms op
        # scores = -scores
        # scores[keep] = -scores[keep]
        # result_mask = torch.zeros(scores.size()).to(scores.device)
        # result_mask[keep] = 1
        # scores = torch.where(result_mask == 1, scores, result_mask)
        boxlist.scores = scores
    else:
        boxlist = boxlist[keep]
    return boxlist

def oks_iou(g, d, a_g, a_d, sigmas=None, in_vis_thre=None):
    """Calculate oks IOU score"""
    if not isinstance(sigmas, np.ndarray):
        sigmas = np.array([.26, .25, .25, .35, .35, .79, .79, .72, .72,
                           .62, .62, 1.07, 1.07, .87, .87, .89, .89]) / 10.0
    vars = (sigmas * 2) ** 2
    xg = g[0::3]
    yg = g[1::3]
    vg = g[2::3]
    ious = np.zeros((d.shape[0]))
    for n_d in range(0, d.shape[0]):
        xd = d[n_d, 0::3]
        yd = d[n_d, 1::3]
        vd = d[n_d, 2::3]
        dx = xd - xg
        dy = yd - yg
        e = (dx ** 2 + dy ** 2) / vars / ((a_g + a_d[n_d]) / 2 + np.spacing(1)) / 2
        if in_vis_thre is not None:
            ind = list(vg > in_vis_thre) and list(vd > in_vis_thre)
            e = e[ind]
        ious[n_d] = np.sum(np.exp(-e)) / e.shape[0] if e.shape[0] != 0 else 0.0
    return ious

def get_close_kpt_num(g, d, dis_thresh=1):
    """Calculate number of close keypoints"""
    xg = g[0::3]
    yg = g[1::3]
    close_keypoints_num = np.zeros((d.shape[0]))
    for n_d in range(0, d.shape[0]):
        xd = d[n_d, 0::3]
        yd = d[n_d, 1::3]
        dx = xd - xg
        dy = yd - yg
        dis = np.sqrt(dx ** 2 + dy ** 2)
        close_keypoints_num[n_d] = (dis <= dis_thresh).sum()
    return close_keypoints_num

def oks_nms(boxlist, thresh, sigmas=None, in_vis_thre=None):
    """
    greedily select boxes with high confidence and overlap with current maximum <= thresh
    rule out overlap >= thresh, overlap = oks
    :param kpts_db
    :param thresh: retain overlap < thresh
    :return: indexes to keep
    """
    if len(boxlist) == 0:
        return boxlist

    kpts = boxlist.pred_keypoints.reshape(len(boxlist), -1).cpu().numpy()
    boxes = boxlist.pred_boxes.tensor
    areas = ((boxes[:, 2] - boxes[:, 0] + 1) * (boxes[:, 3] - boxes[:, 1] + 1)).cpu().numpy()

    # scores = boxlist.scores
    # order = scores.argsort()[::-1]
    order = np.arange(0, len(boxlist))

    keep = []
    while len(order) > 0:
        i = order[0]
        keep.append(i)

        oks_ovr = oks_iou(kpts[i], kpts[order[1:]], areas[i], areas[order[1:]], sigmas, in_vis_thre)

        inds = np.where(oks_ovr <= thresh)[0]
        order = order[inds + 1]

    return boxlist[keep]

def close_kpt_nms(boxlist, dis_thresh=1.0, close_kpt_thresh=5):
    """
    greedily select boxes with high confidence and overlap with current maximum <= thresh
    rule out overlap >= thresh, overlap = oks
    :param kpts_db
    :param thresh: retain overlap < thresh
    :return: indexes to keep
    """
    if len(boxlist) == 0:
        return boxlist

    kpts = boxlist.pred_keypoints.reshape(len(boxlist), -1).cpu().numpy()

    # scores = boxlist.scores
    # order = scores.argsort()[::-1]
    order = np.arange(0, len(boxlist))

    keep = []
    while len(order) > 0:
        i = order[0]
        keep.append(i)

        close_kpt_num = get_close_kpt_num(kpts[i], kpts[order[1:]], dis_thresh)

        inds = np.where(close_kpt_num <= close_kpt_thresh)[0]
        order = order[inds + 1]

    return boxlist[keep]



# adapted from https://github.com/facebookresearch/detectron2/blob/master/detectron2/data/datasets/coco.py
import contextlib
import copy
import datetime
import time
import io
import json
import logging
import os
from PIL import Image
import numpy as np
import portalocker


from ..detection.detection_utils import try_import_pycocotools
from ..structures import Boxes, BoxMode, PolygonMasks
from .catalog import DatasetCatalog, MetadataCatalog


logger = logging.getLogger(__name__)

__all__ = ["load_coco_json", "load_sem_seg", "convert_to_coco_json"]


def load_coco_json(json_file, image_root, dataset_name=None, extra_annotation_keys=None):
    """
    Load a json file with COCO's instances annotation format.
    Currently supports instance detection, instance segmentation,
    and person keypoints annotations.

    Args:
        json_file (str): full path to the json file in COCO instances annotation format.
        image_root (str or path-like): the directory where the images in this json file exists.
        dataset_name (str): the name of the dataset (e.g., coco_2017_train).
            If provided, this function will also put "thing_classes" into
            the metadata associated with this dataset.
        extra_annotation_keys (list[str]): list of per-annotation keys that should also be
            loaded into the dataset dict (besides "iscrowd", "bbox", "keypoints",
            "category_id", "segmentation"). The values for these keys will be returned as-is.
            For example, the densepose annotations are loaded in this way.

    Returns:
        list[dict]: a list of dicts in Detectron2 standard dataset dicts format. (See
        `Using Custom Datasets </tutorials/datasets.html>`_ )

    Notes:
        1. This function does not read the image files.
           The results do not have the "image" field.
    """
    try_import_pycocotools()
    import pycocotools.mask as mask_util
    from pycocotools.coco import COCO

    tic = time.perf_counter()
    json_file = os.fspath(json_file)
    with contextlib.redirect_stdout(io.StringIO()):
        coco_api = COCO(json_file)
    if time.perf_counter() - tic > 1:
        logger.info("Loading {} takes {:.2f} seconds.".format(json_file, time.perf_counter() - tic))

    id_map = None
    if dataset_name is not None:
        meta = MetadataCatalog.get(dataset_name)
        cat_ids = sorted(coco_api.getCatIds())
        cats = coco_api.loadCats(cat_ids)
        # The categories in a custom json file may not be sorted.
        thing_classes = [c["name"] for c in sorted(cats, key=lambda x: x["id"])]
        meta.thing_classes = thing_classes

        # In COCO, certain category ids are artificially removed,
        # and by convention they are always ignored.
        # We deal with COCO's id issue and translate
        # the category ids to contiguous ids in [0, 80).

        # It works by looking at the "categories" field in the json, therefore
        # if users' own json also have incontiguous ids, we'll
        # apply this mapping as well but print a warning.
        if not (min(cat_ids) == 1 and max(cat_ids) == len(cat_ids)):
            if "coco" not in dataset_name:
                logger.warning(
                    """
Category ids in annotations are not in [1, #categories]! We'll apply a mapping for you.
"""
                )
        id_map = {v: i for i, v in enumerate(cat_ids)}
        meta.thing_dataset_id_to_contiguous_id = id_map

    # sort indices for reproducible results
    img_ids = sorted(coco_api.imgs.keys())
    # imgs is a list of dicts, each looks something like:
    # {'license': 4,
    #  'url': 'http://farm6.staticflickr.com/5454/9413846304_881d5e5c3b_z.jpg',
    #  'file_name': 'COCO_val2014_000000001268.jpg',
    #  'height': 427,
    #  'width': 640,
    #  'date_captured': '2013-11-17 05:57:24',
    #  'id': 1268}
    imgs = coco_api.loadImgs(img_ids)
    # anns is a list[list[dict]], where each dict is an annotation
    # record for an object. The inner list enumerates the objects in an image
    # and the outer list enumerates over images. Example of anns[0]:
    # [{'segmentation': [[192.81,
    #     247.09,
    #     ...
    #     219.03,
    #     249.06]],
    #   'area': 1035.749,
    #   'iscrowd': 0,
    #   'image_id': 1268,
    #   'bbox': [192.81, 224.8, 74.73, 33.43],
    #   'category_id': 16,
    #   'id': 42986},
    #  ...]
    anns = [coco_api.imgToAnns[img_id] for img_id in img_ids]

    if "minival" not in json_file:
        # The popular valminusminival & minival annotations for COCO2014 contain this bug.
        # However the ratio of buggy annotations there is tiny and does not affect accuracy.
        # Therefore we explicitly white-list them.
        ann_ids = [ann["id"] for anns_per_image in anns for ann in anns_per_image]
        assert len(set(ann_ids)) == len(ann_ids), "Annotation ids in '{}' are not unique!".format(
            json_file
        )

    imgs_anns = list(zip(imgs, anns))

    logger.info("Loaded {} images in COCO format from {}".format(len(imgs_anns), json_file))

    dataset_dicts = []

    ann_keys = ["iscrowd", "bbox", "keypoints", "category_id"] + (extra_annotation_keys or [])

    num_instances_without_valid_segmentation = 0

    for (img_dict, anno_dict_list) in imgs_anns:
        record = {}
        record["file_name"] = os.path.join(image_root, img_dict["file_name"])
        record["height"] = img_dict["height"]
        record["width"] = img_dict["width"]
        image_id = record["image_id"] = img_dict["id"]

        objs = []
        for anno in anno_dict_list:
            # Check that the image_id in this annotation is the same as
            # the image_id we're looking at.
            # This fails only when the data parsing logic or the annotation file is buggy.

            # The original COCO valminusminival2014 & minival2014 annotation files
            # actually contains bugs that, together with certain ways of using COCO API,
            # can trigger this assertion.
            assert anno["image_id"] == image_id

            assert anno.get("ignore", 0) == 0, '"ignore" in COCO json file is not supported.'

            obj = {key: anno[key] for key in ann_keys if key in anno}

            segm = anno.get("segmentation", None)
            if segm:  # either list[list[float]] or dict(RLE)
                if isinstance(segm, dict):
                    if isinstance(segm["counts"], list):
                        # convert to compressed RLE
                        segm = mask_util.frPyObjects(segm, *segm["size"])
                else:
                    # filter out invalid polygons (< 3 points)
                    segm = [poly for poly in segm if len(poly) % 2 == 0 and len(poly) >= 6]
                    if len(segm) == 0:
                        num_instances_without_valid_segmentation += 1
                        continue  # ignore this instance
                obj["segmentation"] = segm

            keypts = anno.get("keypoints", None)
            if keypts:  # list[int]
                for idx, v in enumerate(keypts):
                    if idx % 3 != 2:
                        # COCO's segmentation coordinates are floating points in [0, H or W],
                        # but keypoint coordinates are integers in [0, H-1 or W-1]
                        # Therefore we assume the coordinates are "pixel indices" and
                        # add 0.5 to convert to floating point coordinates.
                        keypts[idx] = v + 0.5
                obj["keypoints"] = keypts

            obj["bbox_mode"] = BoxMode.XYWH_ABS
            if id_map:
                obj["category_id"] = id_map[obj["category_id"]]
            objs.append(obj)
        record["annotations"] = objs
        dataset_dicts.append(record)

    if num_instances_without_valid_segmentation > 0:
        logger.warning(
            "Filtered out {} instances without valid segmentation. ".format(
                num_instances_without_valid_segmentation
            )
            + "There might be issues in your dataset generation process. "
            "A valid polygon should be a list[float] with even length >= 6."
        )
    return dataset_dicts


def load_sem_seg(gt_root, image_root, gt_ext="png", image_ext="jpg"):
    """
    Load semantic segmentation datasets. All files under "gt_root" with "gt_ext" extension are
    treated as ground truth annotations and all files under "image_root" with "image_ext" extension
    as input images. Ground truth and input images are matched using file paths relative to
    "gt_root" and "image_root" respectively without taking into account file extensions.
    This works for COCO as well as some other datasets.

    Args:
        gt_root (str): full path to ground truth semantic segmentation files. Semantic segmentation
            annotations are stored as images with integer values in pixels that represent
            corresponding semantic labels.
        image_root (str): the directory where the input images are.
        gt_ext (str): file extension for ground truth annotations.
        image_ext (str): file extension for input images.

    Returns:
        list[dict]:
            a list of dicts in detectron2 standard format without instance-level
            annotation.

    Notes:
        1. This function does not read the image and ground truth files.
           The results do not have the "image" and "sem_seg" fields.
    """

    # We match input images with ground truth based on their relative filepaths (without file
    # extensions) starting from 'image_root' and 'gt_root' respectively.
    def file2id(folder_path, file_path):
        # extract relative path starting from `folder_path`
        image_id = os.path.normpath(os.path.relpath(file_path, start=folder_path))
        # remove file extension
        image_id = os.path.splitext(image_id)[0]
        return image_id

    input_files = sorted(
        (os.path.join(image_root, f) for f in os.listdir(image_root) if f.endswith(image_ext)),
        key=lambda file_path: file2id(image_root, file_path),
    )
    gt_files = sorted(
        (os.path.join(gt_root, f) for f in os.listdir(gt_root) if f.endswith(gt_ext)),
        key=lambda file_path: file2id(gt_root, file_path),
    )

    assert len(gt_files) > 0, "No annotations found in {}.".format(gt_root)

    # Use the intersection, so that val2017_100 annotations can run smoothly with val2017 images
    if len(input_files) != len(gt_files):
        logger.warning(
            "Directory {} and {} has {} and {} files, respectively.".format(
                image_root, gt_root, len(input_files), len(gt_files)
            )
        )
        input_basenames = [os.path.basename(f)[: -len(image_ext)] for f in input_files]
        gt_basenames = [os.path.basename(f)[: -len(gt_ext)] for f in gt_files]
        intersect = list(set(input_basenames) & set(gt_basenames))
        # sort, otherwise each worker may obtain a list[dict] in different order
        intersect = sorted(intersect)
        logger.warning("Will use their intersection of {} files.".format(len(intersect)))
        input_files = [os.path.join(image_root, f + image_ext) for f in intersect]
        gt_files = [os.path.join(gt_root, f + gt_ext) for f in intersect]

    logger.info(
        "Loaded {} images with semantic segmentation from {}".format(len(input_files), image_root)
    )

    dataset_dicts = []
    for (img_path, gt_path) in zip(input_files, gt_files):
        record = {}
        record["file_name"] = img_path
        record["sem_seg_file_name"] = gt_path
        dataset_dicts.append(record)

    return dataset_dicts


def convert_to_coco_dict(dataset_name):
    """
    Convert an instance detection/segmentation or keypoint detection dataset
    in detectron2's standard format into COCO json format.

    Generic dataset description can be found here:
    https://detectron2.readthedocs.io/tutorials/datasets.html#register-a-dataset

    COCO data format description can be found here:
    http://cocodataset.org/#format-data

    Args:
        dataset_name (str):
            name of the source dataset
            Must be registered in DatastCatalog and in detectron2's standard format.
            Must have corresponding metadata "thing_classes"
    Returns:
        coco_dict: serializable dict in COCO json format
    """
    try_import_pycocotools()
    import pycocotools.mask as mask_util
    dataset_dicts = DatasetCatalog.get(dataset_name)
    metadata = MetadataCatalog.get(dataset_name)

    # unmap the category mapping ids for COCO
    if hasattr(metadata, "thing_dataset_id_to_contiguous_id"):
        reverse_id_mapping = {v: k for k, v in metadata.thing_dataset_id_to_contiguous_id.items()}
        reverse_id_mapper = lambda contiguous_id: reverse_id_mapping[contiguous_id]  # noqa
    else:
        reverse_id_mapper = lambda contiguous_id: contiguous_id  # noqa

    categories = [
        {"id": reverse_id_mapper(id), "name": name}
        for id, name in enumerate(metadata.thing_classes)
    ]

    logger.info("Converting dataset dicts into COCO format")
    coco_images = []
    coco_annotations = []

    for image_id, image_dict in enumerate(dataset_dicts):
        coco_image = {
            "id": image_dict.get("image_id", image_id),
            "width": image_dict["width"],
            "height": image_dict["height"],
            "file_name": image_dict["file_name"],
        }
        coco_images.append(coco_image)

        anns_per_image = image_dict.get("annotations", [])
        for annotation in anns_per_image:
            # create a new dict with only COCO fields
            coco_annotation = {}

            # COCO requirement: XYWH box format
            bbox = annotation["bbox"]
            bbox_mode = annotation["bbox_mode"]
            bbox = BoxMode.convert(bbox, bbox_mode, BoxMode.XYWH_ABS)

            # COCO requirement: instance area
            if "segmentation" in annotation:
                # Computing areas for instances by counting the pixels
                segmentation = annotation["segmentation"]
                # TODO: check segmentation type: RLE, BinaryMask or Polygon
                if isinstance(segmentation, list):
                    polygons = PolygonMasks([segmentation])
                    area = polygons.area()[0].item()
                elif isinstance(segmentation, dict):  # RLE
                    area = mask_util.area(segmentation).item()
                else:
                    raise TypeError(f"Unknown segmentation type {type(segmentation)}!")
            else:
                # Computing areas using bounding boxes
                bbox_xy = BoxMode.convert(bbox, BoxMode.XYWH_ABS, BoxMode.XYXY_ABS)
                area = Boxes([bbox_xy]).area()[0].item()

            if "keypoints" in annotation:
                keypoints = annotation["keypoints"]  # list[int]
                for idx, v in enumerate(keypoints):
                    if idx % 3 != 2:
                        # COCO's segmentation coordinates are floating points in [0, H or W],
                        # but keypoint coordinates are integers in [0, H-1 or W-1]
                        # For COCO format consistency we substract 0.5
                        # https://github.com/facebookresearch/detectron2/pull/175#issuecomment-551202163
                        keypoints[idx] = v - 0.5
                if "num_keypoints" in annotation:
                    num_keypoints = annotation["num_keypoints"]
                else:
                    num_keypoints = sum(kp > 0 for kp in keypoints[2::3])

            # COCO requirement:
            #   linking annotations to images
            #   "id" field must start with 1
            coco_annotation["id"] = len(coco_annotations) + 1
            coco_annotation["image_id"] = coco_image["id"]
            coco_annotation["bbox"] = [round(float(x), 3) for x in bbox]
            coco_annotation["area"] = float(area)
            coco_annotation["iscrowd"] = annotation.get("iscrowd", 0)
            coco_annotation["category_id"] = reverse_id_mapper(annotation["category_id"])

            # Add optional fields
            if "keypoints" in annotation:
                coco_annotation["keypoints"] = keypoints
                coco_annotation["num_keypoints"] = num_keypoints

            if "segmentation" in annotation:
                seg = coco_annotation["segmentation"] = annotation["segmentation"]
                if isinstance(seg, dict):  # RLE
                    counts = seg["counts"]
                    if not isinstance(counts, str):
                        # make it json-serializable
                        seg["counts"] = counts.decode("ascii")

            coco_annotations.append(coco_annotation)

    logger.info(
        "Conversion finished, "
        f"#images: {len(coco_images)}, #annotations: {len(coco_annotations)}"
    )

    info = {
        "date_created": str(datetime.datetime.now()),
        "description": "Automatically generated COCO json file for Detectron2.",
    }
    coco_dict = {"info": info, "images": coco_images, "categories": categories, "licenses": None}
    if len(coco_annotations) > 0:
        coco_dict["annotations"] = coco_annotations
    return coco_dict


def convert_to_coco_json(dataset_name, output_file, allow_cached=True):
    """
    Converts dataset into COCO format and saves it to a json file.
    dataset_name must be registered in DatasetCatalog and in detectron2's standard format.

    Args:
        dataset_name:
            reference from the config file to the catalogs
            must be registered in DatasetCatalog and in detectron2's standard format
        output_file: path of json file that will be saved to
        allow_cached: if json file is already present then skip conversion
    """

    # TODO: The dataset or the conversion script *may* change,
    # a checksum would be useful for validating the cached data

    os.makedirs(os.path.dirname(output_file), exist_ok=True)
    with portalocker.Lock(output_file):
        if os.path.isfile(output_file) and allow_cached:
            logger.warning(
                f"Using previously cached COCO format annotations at '{output_file}'. "
                "You need to clear the cache file if your dataset has been modified."
            )
        else:
            logger.info(f"Converting annotations of dataset '{dataset_name}' to COCO format ...)")
            coco_dict = convert_to_coco_dict(dataset_name)

            logger.info(f"Caching COCO format annotations at '{output_file}' ...")
            with open(output_file, "w") as f:
                json.dump(coco_dict, f)

def register_coco_instances(name, metadata, json_file, image_root):
    """
    Register a dataset in COCO's json annotation format for
    instance detection, instance segmentation and keypoint detection.
    (i.e., Type 1 and 2 in http://cocodataset.org/#format-data.
    `instances*.json` and `person_keypoints*.json` in the dataset).

    This is an example of how to register a new dataset.
    You can do something similar to this function, to register new datasets.

    Args:
        name (str): the name that identifies a dataset, e.g. "coco_2014_train".
        metadata (dict): extra metadata associated with this dataset.  You can
            leave it as an empty dict.
        json_file (str): path to the json instance annotation file.
        image_root (str or path-like): directory which contains all the images.
    """
    assert isinstance(name, str), name
    assert isinstance(json_file, (str, os.PathLike)), json_file
    assert isinstance(image_root, (str, os.PathLike)), image_root
    # 1. register a function which returns dicts
    DatasetCatalog.register(name, lambda: load_coco_json(json_file, image_root, name))

    # 2. Optionally, add metadata about this dataset,
    # since they might be useful in evaluation, visualization or logging
    MetadataCatalog.get(name).set(
        json_file=json_file, image_root=image_root, evaluator_type="coco", **metadata
    )

def register_coco_panoptic_separated(name, metadata, image_root, panoptic_root,
                                     panoptic_json, sem_seg_root, instances_json):
    """
    Register a COCO panoptic segmentation dataset named `name`.
    The annotations in this registered dataset will contain both instance annotations and
    semantic annotations, each with its own contiguous ids. Hence it's called "separated".

    It follows the setting used by the PanopticFPN paper:

    1. The instance annotations directly come from polygons in the COCO
       instances annotation task, rather than from the masks in the COCO panoptic annotations.

       The two format have small differences:
       Polygons in the instance annotations may have overlaps.
       The mask annotations are produced by labeling the overlapped polygons
       with depth ordering.

    2. The semantic annotations are converted from panoptic annotations, where
       all "things" are assigned a semantic id of 0.
       All semantic categories will therefore have ids in contiguous
       range [1, #stuff_categories].

    This function will also register a pure semantic segmentation dataset
    named ``name + '_stuffonly'``.

    Args:
        name (str): the name that identifies a dataset,
            e.g. "coco_2017_train_panoptic"
        metadata (dict): extra metadata associated with this dataset.
        image_root (str): directory which contains all the images
        panoptic_root (str): directory which contains panoptic annotation images
        panoptic_json (str): path to the json panoptic annotation file
        sem_seg_root (str): directory which contains all the ground truth segmentation annotations.
        instances_json (str): path to the json instance annotation file
    """
    panoptic_name = name + "_separated"
    DatasetCatalog.register(
        panoptic_name,
        lambda: merge_to_panoptic(
            load_coco_json(instances_json, image_root, panoptic_name),
            load_sem_seg(sem_seg_root, image_root),
        ),
    )
    MetadataCatalog.get(panoptic_name).set(
        panoptic_root=panoptic_root,
        image_root=image_root,
        panoptic_json=panoptic_json,
        sem_seg_root=sem_seg_root,
        json_file=instances_json,  # TODO rename
        evaluator_type="coco_panoptic_seg",
        **metadata
    )

    semantic_name = name + "_stuffonly"
    DatasetCatalog.register(semantic_name, lambda: load_sem_seg(sem_seg_root, image_root))
    MetadataCatalog.get(semantic_name).set(
        sem_seg_root=sem_seg_root, image_root=image_root, evaluator_type="sem_seg", **metadata
    )


def merge_to_panoptic(detection_dicts, sem_seg_dicts):
    """
    Create dataset dicts for panoptic segmentation, by
    merging two dicts using "file_name" field to match their entries.

    Args:
        detection_dicts (list[dict]): lists of dicts for object detection or instance segmentation.
        sem_seg_dicts (list[dict]): lists of dicts for semantic segmentation.

    Returns:
        list[dict] (one per input image): Each dict contains all (key, value) pairs from dicts in
            both detection_dicts and sem_seg_dicts that correspond to the same image.
            The function assumes that the same key in different dicts has the same value.
    """
    results = []
    sem_seg_file_to_entry = {x["file_name"]: x for x in sem_seg_dicts}
    assert len(sem_seg_file_to_entry) > 0

    for det_dict in detection_dicts:
        dic = copy.copy(det_dict)
        dic.update(sem_seg_file_to_entry[dic["file_name"]])
        results.append(dic)
    return results



"""Learning rate policy
Gradually warm-up(increasing) learning rate for pytorch's optimizer.
Proposed in 'Accurate, Large Minibatch SGD: Training ImageNet in 1 Hour'.
Code adapted from https://github.com/ildoonet/pytorch-gradual-warmup-lr
"""
# pylint: disable=missing-function-docstring, line-too-long, inconsistent-return-statements, redefined-builtin,missing-class-docstring
from typing import List
import torch
from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.lr_scheduler import ReduceLROnPlateau


class GradualWarmupScheduler(_LRScheduler):
    """ Gradually warm-up(increasing) learning rate in optimizer.
    Proposed in 'Accurate, Large Minibatch SGD: Training ImageNet in 1 Hour'.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        multiplier: target learning rate = base lr * multiplier if multiplier > 1.0. if multiplier = 1.0, lr starts from 0 and ends up with the base_lr.
        total_epoch: target learning rate is reached at total_epoch, gradually
        after_scheduler: after target_epoch, use this scheduler(eg. ReduceLROnPlateau)
    """
    def __init__(self,
                 optimizer,
                 multiplier,
                 total_epoch,
                 after_scheduler=None):
        self.multiplier = multiplier
        if self.multiplier < 1.:
            raise ValueError(
                'multiplier should be greater thant or equal to 1.')
        self.total_epoch = total_epoch
        self.after_scheduler = after_scheduler
        self.finished = False
        super(GradualWarmupScheduler, self).__init__(optimizer)

    def get_lr(self):
        if self.last_epoch > self.total_epoch:
            if self.after_scheduler:
                if not self.finished:
                    self.after_scheduler.base_lrs = [
                        base_lr * self.multiplier for base_lr in self.base_lrs
                    ]
                    self.finished = True
                return self.after_scheduler.get_last_lr()
            return [base_lr * self.multiplier for base_lr in self.base_lrs]

        if self.multiplier == 1.0:
            return [
                base_lr * (float(self.last_epoch) / self.total_epoch)
                for base_lr in self.base_lrs
            ]
        else:
            return [
                base_lr *
                ((self.multiplier - 1.) * self.last_epoch / self.total_epoch +
                 1.) for base_lr in self.base_lrs
            ]

    def step_ReduceLROnPlateau(self, metrics, epoch=None):
        # pylint: disable=access-member-before-definition
        if epoch is None:
            epoch = self.last_epoch + 1
        # ReduceLROnPlateau is called at the end of epoch, whereas others are called at beginning
        self.last_epoch = epoch if epoch != 0 else 1
        if self.last_epoch <= self.total_epoch:
            warmup_lr = [
                base_lr *
                ((self.multiplier - 1.) * self.last_epoch / self.total_epoch +
                 1.) for base_lr in self.base_lrs
            ]
            for param_group, lr in zip(self.optimizer.param_groups, warmup_lr):
                param_group['lr'] = lr
        else:
            if epoch is None:
                self.after_scheduler.step(metrics, None)
            else:
                self.after_scheduler.step(metrics, epoch - self.total_epoch)

    def step(self, epoch=None, metrics=None):
        if not isinstance(self.after_scheduler, ReduceLROnPlateau):
            if self.finished and self.after_scheduler:
                if epoch is None:
                    self.after_scheduler.step(None)
                else:
                    self.after_scheduler.step(epoch - self.total_epoch)
                self._last_lr = self.after_scheduler.get_last_lr()
            else:
                return super(GradualWarmupScheduler, self).step(epoch)
        else:
            self.step_ReduceLROnPlateau(metrics, epoch)


class ReduceLROnPlateauWarmup(ReduceLROnPlateau):
    """ Reduce LR on plateau """
    def __init__(self, optimizer, warmup_epochs, **kwargs):
        self.warmup_epochs = warmup_epochs
        super().__init__(optimizer, **kwargs)
        self.last_epoch = 0
        self.base_lrs = []
        for group in optimizer.param_groups:
            self.base_lrs.append(group["lr"])
        self.step_rop(self.mode_worse, False, None)

    def step_rop(self, metrics, do_eval, epoch=None):
        assert epoch is None
        epoch = self.last_epoch + 1
        if epoch <= self.warmup_epochs:
            factor = epoch / self.warmup_epochs
            self.last_epoch = epoch
            for i, param_group in enumerate(self.optimizer.param_groups):
                param_group['lr'] = self.base_lrs[i] * factor
        elif not do_eval:
            pass
        else:
            super().step(metrics, epoch=epoch)

class WarmupLinearLR(torch.optim.lr_scheduler._LRScheduler):
    def __init__(self,
                 optimizer: torch.optim.Optimizer,
                 max_iter: int = 90000,
                 warmup_factor: float = 0.001,
                 warmup_iters: int = 1000,
                 warmup_method: str = "linear",
                 last_epoch: int = -1):
        self.warmup_factor = warmup_factor
        self.warmup_iters = warmup_iters
        self.warmup_method = warmup_method
        self.max_iter = max_iter
        super().__init__(optimizer, last_epoch)

    def get_lr(self) -> List[float]:
        warmup_factor = _get_warmup_factor_at_iter(
            self.warmup_method, self.last_epoch, self.warmup_iters, self.warmup_factor
        )
        return [base_lr * warmup_factor * (1 - self.last_epoch / self.max_iter)
                for base_lr in self.base_lrs]

    def _compute_values(self) -> List[float]:
        # The new interface
        return self.get_lr()

def _get_warmup_factor_at_iter(
        method: str, iter: int, warmup_iters: int, warmup_factor: float) -> float:
    """
    Return the learning rate warmup factor at a specific iteration.
    See :paper:`ImageNet in 1h` for more details.

    Args:
        method (str): warmup method; either "constant" or "linear".
        iter (int): iteration at which to calculate the warmup factor.
        warmup_iters (int): the number of warmup iterations.
        warmup_factor (float): the base warmup factor (the meaning changes according
            to the method used).

    Returns:
        float: the effective warmup factor at the given iteration.
    """
    if iter >= warmup_iters:
        return 1.0

    if method == "constant":
        return warmup_factor
    elif method == "linear":
        alpha = iter / warmup_iters
        return warmup_factor * (1 - alpha) + alpha
    else:
        raise ValueError("Unknown warmup method: {}".format(method))


def build_lr_scheduler(cfg, optimizer):
    """
    Build a LR scheduler from config.
    """
    name = cfg.CONFIG.TRAIN.ITER_LR_SCHEDULER_NAME
    if name == "WarmupMultiStepLR":
        return WarmupMultiStepLR(
            optimizer,
            cfg.CONFIG.TRAIN.ITER_LR_STEPS,
            cfg.CONFIG.TRAIN.STEP,
            warmup_factor=cfg.CONFIG.TRAIN.ITER_BASED_WARMUP_FACTOR,
            warmup_iters=cfg.CONFIG.TRAIN.ITER_BASED_WARMUP_ITERS,
            warmup_method=cfg.CONFIG.TRAIN.ITER_BASED_WARMUP_METHOD,
        )
    elif name == "WarmupCosineLR":
        return WarmupCosineLR(
            optimizer,
            cfg.CONFIG.TRAIN.ITER_NUM,
            warmup_factor=cfg.CONFIG.TRAIN.ITER_BASED_WARMUP_FACTOR,
            warmup_iters=cfg.CONFIG.TRAIN.ITER_BASED_WARMUP_ITERS,
            warmup_method=cfg.CONFIG.TRAIN.ITER_BASED_WARMUP_METHOD,
        )
    elif name == 'WarmupLinearLR':
        return WarmupLinearLR(
            optimizer=optimizer,
            max_iter=cfg.CONFIG.TRAIN.ITER_NUM,
            warmup_factor=cfg.CONFIG.TRAIN.ITER_BASED_WARMUP_FACTOR,
            warmup_iters=cfg.CONFIG.TRAIN.ITER_BASED_WARMUP_ITERS,
            warmup_method=cfg.CONFIG.TRAIN.ITER_BASED_WARMUP_METHOD,
        )
    else:
        raise ValueError("Unknown LR scheduler: {}".format(name))



"""Default setting in training/testing for coot"""
from yacs.config import CfgNode as CN


_C = CN()

# ---------------------------------------------------------------------------- #
# Distributed DataParallel setting: DDP_CONFIG
# ---------------------------------------------------------------------------- #

_C.DDP_CONFIG = CN(new_allowed=False)
# Number of nodes for distributed training
_C.DDP_CONFIG.WORLD_SIZE = 1
# Node rank for distributed training
_C.DDP_CONFIG.WORLD_RANK = 0
# Number of GPUs to use
_C.DDP_CONFIG.GPU_WORLD_SIZE = 8
# GPU rank for distributed training
_C.DDP_CONFIG.GPU_WORLD_RANK = 0
# Master node
_C.DDP_CONFIG.DIST_URL = 'tcp://127.0.0.1:10001'
# A list of IP addresses for each node
_C.DDP_CONFIG.WOLRD_URLS = ['127.0.0.1']
# Whether to turn on automatic ranking match.
_C.DDP_CONFIG.AUTO_RANK_MATCH = True
# distributed backend
_C.DDP_CONFIG.DIST_BACKEND = 'nccl'
# Current GPU id.
_C.DDP_CONFIG.GPU = 0
# Whether to use distributed training or simply use dataparallel.
_C.DDP_CONFIG.DISTRIBUTED = True

# ---------------------------------------------------------------------------- #
# Standard training/testing setting: CONFIG
# ---------------------------------------------------------------------------- #

_C.CONFIG = CN(new_allowed=True)

_C.CONFIG.TRAIN = CN(new_allowed=True)
# Maximal number of epochs.
_C.CONFIG.TRAIN.EPOCH_NUM = 196
# Per GPU mini-batch size.
_C.CONFIG.TRAIN.BATCH_SIZE = 8
# Base learning rate.
_C.CONFIG.TRAIN.LR = 0.01
# Momentum.
_C.CONFIG.TRAIN.MOMENTUM = 0.9
# Adam Beta 2
_C.CONFIG.TRAIN.ADAM_BETA2 = 0.98
# Adam eps
_C.CONFIG.TRAIN.ADAM_EPS = 1.5e-09
# L2 regularization.
_C.CONFIG.TRAIN.W_DECAY = 1e-4
# Learning rate policy
_C.CONFIG.TRAIN.LR_POLICY: 'Cosine'
# Steps for multistep learning rate policy
_C.CONFIG.TRAIN.LR_MILESTONE = [40, 80]
# Exponential decay factor.
_C.CONFIG.TRAIN.STEP = 0.1
# Use warm up scheduler or not.
_C.CONFIG.TRAIN.USE_WARMUP: False
# Gradually warm up the SOLVER.BASE_LR over this number of epochs.
_C.CONFIG.TRAIN.WARMUP_EPOCHS: 34
# The start learning rate of the warm up.
_C.CONFIG.TRAIN.WARMUP_START_LR: 0.01
# The end learning rate of the warm up.
_C.CONFIG.TRAIN.WARMUP_END_LR: 0.1
# Resume training from a specific epoch. Set to -1 means train from beginning.
_C.CONFIG.TRAIN.RESUME_EPOCH: -1

# Whether to use multigrid training to speed up.
_C.CONFIG.TRAIN.MULTIGRID = CN(new_allowed=True)
_C.CONFIG.TRAIN.MULTIGRID.USE_LONG_CYCLE = False
_C.CONFIG.TRAIN.MULTIGRID.USE_SHORT_CYCLE = False
_C.CONFIG.TRAIN.MULTIGRID.LONG_CYCLE_EPOCH = [10, 20, 30]

_C.CONFIG.VAL = CN(new_allowed=True)
# Evaluate model on test data every eval period epochs.
_C.CONFIG.VAL.FREQ = 2
# Per GPU mini-batch size.
_C.CONFIG.VAL.BATCH_SIZE = 8


_C.CONFIG.INFERENCE = CN(new_allowed=True)
# Whether to extract features or make predictions.
# If set to True, only features will be returned.
_C.CONFIG.INFERENCE.FEAT = False


_C.CONFIG.DATA = CN(new_allowed=True)

# Paths of annotation files and actual data
_C.CONFIG.DATA.TRAIN_ANNO_PATH = ''
_C.CONFIG.DATA.TRAIN_DATA_PATH = ''
_C.CONFIG.DATA.VAL_ANNO_PATH = ''
_C.CONFIG.DATA.VAL_DATA_PATH = ''
# The number of classes to predict for the model.
_C.CONFIG.DATA.NUM_CLASSES = 400
# The number of frames of the input clip.
_C.CONFIG.DATA.CLIP_LEN = 16
# The video sampling rate of the input clip.
_C.CONFIG.DATA.FRAME_RATE = 2
# Whether to keep aspect ratio when resizing input
_C.CONFIG.DATA.KEEP_ASPECT_RATIO = False
# Temporal segment setting for training video action recognition models.
_C.CONFIG.DATA.NUM_SEGMENT = 1
_C.CONFIG.DATA.NUM_CROP = 1
# Multi-view evaluation for video action recognition models.
# Usually for 2D models, it is 25 segments with 10 crops.
# For 3D models, it is 10 segments with 3 crops.
# Number of clips to sample from a video uniformly for aggregating the
# prediction results.
_C.CONFIG.DATA.TEST_NUM_SEGMENT = 10
# Number of crops to sample from a frame spatially for aggregating the
# prediction results.
_C.CONFIG.DATA.TEST_NUM_CROP = 3
# The spatial crop size for training.
_C.CONFIG.DATA.CROP_SIZE = 224
# Size of the smallest side of the image during testing.
_C.CONFIG.DATA.SHORT_SIDE_SIZE = 256
# Pre-defined height for resizing input video frames.
_C.CONFIG.DATA.NEW_HEIGHT = 256
# Pre-defined width for resizing input video frames.
_C.CONFIG.DATA.NEW_WIDTH = 340
_C.CONFIG.DATA.NUM_WORKERS = 0

_C.CONFIG.MODEL = CN(new_allowed=True)
# Model architecture. You can find available models in the model zoo.
_C.CONFIG.MODEL.NAME = ''
# Whether to load a checkpoint file. If True, please set the following PRETRAINED_PATH.
_C.CONFIG.MODEL.LOAD = False
# Path (a file path, or URL) to a checkpoint file to be loaded to the model.
_C.CONFIG.MODEL.PRETRAINED_PATH = ''
# Whether to use the trained weights in the model zoo.
_C.CONFIG.MODEL.PRETRAINED = False
# Whether to use pretrained backbone network. Usually this is set to True.
_C.CONFIG.MODEL.PRETRAINED_BASE = True
# BN options
_C.CONFIG.MODEL.BN_EVAL = False
_C.CONFIG.MODEL.PARTIAL_BN = False
_C.CONFIG.MODEL.BN_FROZEN = False
_C.CONFIG.MODEL.USE_AFFINE = False


_C.CONFIG.LOG = CN(new_allowed=True)
# Base directory where all output files are written
_C.CONFIG.LOG.BASE_PATH = ''
# Pre-defined name for each experiment.
# If set to 'use_time', the start time will be appended to the directory name.
_C.CONFIG.LOG.EXP_NAME = 'use_time'
# Directory where training logs are written
_C.CONFIG.LOG.LOG_DIR = 'tb_log'
# Directory where checkpoints are written
_C.CONFIG.LOG.SAVE_DIR = 'checkpoints'
# Directory where testing logs are written
_C.CONFIG.LOG.EVAL_DIR = ''
# Save a checkpoint after every this number of epochs
_C.CONFIG.LOG.SAVE_FREQ = 1
# Display the training log after every this number of iterations
_C.CONFIG.LOG.DISPLAY_FREQ = 1


# ---------------------------------------------------------------------------- #
# Video-Language: COOT CONFIG
# ---------------------------------------------------------------------------- #

_C.CONFIG.COOT_DATA = CN(new_allowed=True)
_C.CONFIG.COOT_MODEL = CN(new_allowed=True)
_C.CONFIG.COOT_MODEL.MODEL_CONFIG = CN(new_allowed=True)
_C.CONFIG.COOT_MODEL.MODEL_CONFIG.VIDEO_POOLER = CN(new_allowed=True)
_C.CONFIG.COOT_MODEL.MODEL_CONFIG.VIDEO_SEQUENCER = CN(new_allowed=True)
_C.CONFIG.COOT_MODEL.MODEL_CONFIG.TEXT_POOLER = CN(new_allowed=True)
_C.CONFIG.COOT_MODEL.MODEL_CONFIG.TEXT_SEQUENCER = CN(new_allowed=True)
_C.CONFIG.COOT_MODEL.MODEL_CONFIG.TEXT_ENCODER = CN(new_allowed=True)



"""GluonCV Utility functions."""
# pylint: disable=wildcard-import,exec-used,wrong-import-position
from __future__ import absolute_import

import types

def import_dummy_module(code, name):
    # create blank module
    module = types.ModuleType(name)
    # populate the module with code
    exec(code, module.__dict__)
    return module

dummy_module = """
def __getattr__(name):
    raise AttributeError(f"gluoncv.utils.{__name__} module requires mxnet which is missing.")
"""


from . import bbox
from . import random
from . import filesystem
try:
    import mxnet
    from . import viz
    from . import metrics
    from . import parallel
    from .lr_scheduler import LRSequential, LRScheduler
    from .export_helper import export_block, export_tvm
    from .sync_loader_helper import split_data, split_and_load
except ImportError:
    viz = import_dummy_module(dummy_module, 'viz')
    metrics = import_dummy_module(dummy_module, 'metrics')
    parallel = import_dummy_module(dummy_module, 'parallel')
    LRSequential, LRScheduler = None, None
    export_block, export_tvm = None, None
    split_data, split_and_load = None, None

from .download import download, check_sha1
from .filesystem import makedirs, try_import_dali, try_import_cv2
from .bbox import bbox_iou
from .block import recursive_visit, set_lr_mult, freeze_bn
from .plot_history import TrainingHistory
from .version import *



"""
optimizer RADAM
"""
import math
from enum import Enum

import torch
from torch.optim.optimizer import Optimizer


class RAdam(Optimizer):
    """RADam optimizer
    """
    def __init__(self,
                 params,
                 lr=1e-3,
                 betas=(0.9, 0.999),
                 eps=1e-8,
                 weight_decay=0,
                 degenerated_to_sgd=True):
        if lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if eps < 0.0:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(
                betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(
                betas[1]))

        self.degenerated_to_sgd = degenerated_to_sgd
        if isinstance(params,
                      (list, tuple)) and len(params) > 0 and isinstance(
                          params[0], dict):
            for param in params:
                if 'betas' in param and (param['betas'][0] != betas[0]
                                         or param['betas'][1] != betas[1]):
                    param['buffer'] = [[None, None, None] for _ in range(10)]
        defaults = dict(lr=lr,
                        betas=betas,
                        eps=eps,
                        weight_decay=weight_decay,
                        buffer=[[None, None, None] for _ in range(10)])
        super(RAdam, self).__init__(params, defaults)

    def __setstate__(self, state):
        "set state"
        super(RAdam, self).__setstate__(state)

    def step(self, closure=None):
        """ step gradiend"""
        loss = None
        if closure is not None:
            loss = closure()
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError(
                        'RAdam does not support sparse gradients')
                p_data_fp32 = p.data.float()
                state = self.state[p]
                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p_data_fp32)
                    state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)
                else:
                    state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                    state['exp_avg_sq'] = state['exp_avg_sq'].type_as(
                        p_data_fp32)
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                state['step'] += 1
                buffered = group['buffer'][int(state['step'] % 10)]
                if state['step'] == buffered[0]:
                    n_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state['step']
                    beta2_t = beta2**state['step']
                    n_sma_max = 2 / (1 - beta2) - 1
                    n_sma = n_sma_max - 2 * state['step'] * beta2_t / (1 -
                                                                       beta2_t)
                    buffered[1] = n_sma
                    if n_sma >= 5:
                        step_size = math.sqrt(
                            (1 - beta2_t) * (n_sma - 4) / (n_sma_max - 4) *
                            (n_sma - 2) / n_sma * n_sma_max /
                            (n_sma_max - 2)) / (1 - beta1**state['step'])
                    elif self.degenerated_to_sgd:
                        step_size = 1.0 / (1 - beta1**state['step'])
                    else:
                        step_size = -1
                    buffered[2] = step_size
                if n_sma >= 5:
                    if group['weight_decay'] != 0:
                        p_data_fp32.add_(-group['weight_decay'] * group['lr'],
                                         p_data_fp32)
                    denom = exp_avg_sq.sqrt().add_(group['eps'])
                    p_data_fp32.addcdiv_(-step_size * group['lr'], exp_avg,
                                         denom)
                    p.data.copy_(p_data_fp32)
                elif step_size > 0:
                    if group['weight_decay'] != 0:
                        p_data_fp32.add_(-group['weight_decay'] * group['lr'],
                                         p_data_fp32)
                    p_data_fp32.add_(-step_size * group['lr'], exp_avg)
                    p.data.copy_(p_data_fp32)
        return loss


class GradientClipType(Enum):
    VALUE = "value"
    NORM = "norm"


def _create_gradient_clipper(cfg):
    """
    Creates gradient clipping closure to clip by value or by norm,
    according to the provided config.
    """
    cfg = cfg.clone()

    def clip_grad_norm(p: _GradientClipperInput):
        torch.nn.utils.clip_grad_norm_(p, cfg.CLIP_VALUE, cfg.NORM_TYPE)

    def clip_grad_value(p: _GradientClipperInput):
        torch.nn.utils.clip_grad_value_(p, cfg.CLIP_VALUE)

    _GRADIENT_CLIP_TYPE_TO_CLIPPER = {
        GradientClipType.VALUE: clip_grad_value,
        GradientClipType.NORM: clip_grad_norm,
    }
    return _GRADIENT_CLIP_TYPE_TO_CLIPPER[GradientClipType(cfg.CLIP_TYPE)]


def _generate_optimizer_class_with_gradient_clipping(
        optimizer_type, gradient_clipper):
    """
    Dynamically creates a new type that inherits the type of a given instance
    and overrides the `step` method to add gradient clipping
    """

    def optimizer_wgc_step(self, closure=None):
        for group in self.param_groups:
            for p in group["params"]:
                gradient_clipper(p)
        super(type(self), self).step(closure)

    OptimizerWithGradientClip = type(
        optimizer_type.__name__ + "WithGradientClip",
        (optimizer_type,),
        {"step": optimizer_wgc_step},
    )
    return OptimizerWithGradientClip

def maybe_add_gradient_clipping(cfg, optimizer):
    """
    If gradient clipping is enabled through config options, wraps the existing
    optimizer instance of some type OptimizerType to become an instance
    of the new dynamically created class OptimizerTypeWithGradientClip
    that inherits OptimizerType and overrides the `step` method to
    include gradient clipping.

    Args:
        cfg: CfgNode
            configuration options
        optimizer: torch.optim.Optimizer
            existing optimizer instance

    Return:
        optimizer: torch.optim.Optimizer
            either the unmodified optimizer instance (if gradient clipping is
            disabled), or the same instance with adjusted __class__ to override
            the `step` method and include gradient clipping
    """
    if not cfg.CONFIG.TRAIN.CLIP_GRADIENTS.ENABLED:
        return optimizer
    grad_clipper = _create_gradient_clipper(cfg.CONFIG.TRAIN.CLIP_GRADIENTS)
    OptimizerWithGradientClip = _generate_optimizer_class_with_gradient_clipping(
        type(optimizer), grad_clipper
    )
    optimizer.__class__ = OptimizerWithGradientClip
    return optimizer



"""
This module provides data loaders and transformers for popular vision datasets.
"""

from .video_cls.dataset_classification import VideoClsDataset
from .video_cls.dataset_classification import build_dataloader, build_dataloader_test
from .video_cls.multigrid_helper import multiGridHelper, MultiGridBatchSampler
from .coot.dataloader import create_datasets, create_loaders
from . import registry



"""YOLO default config"""
# pylint: disable=unused-variable,missing-function-docstring,bad-whitespace,missing-class-docstring
from typing import Union, Tuple
from autocfg import dataclass, field


@dataclass
class YOLOv3:
    # Base network name which serves as feature extraction base.
    base_network : str = 'darknet53'
    # List of convolution layer channels which is going to be appended to the
    # base network feature extractor. If `name` is `None`, this is ignored.
    filters : Union[Tuple, list] = (512, 256, 128)
    # The anchor setting.
    anchors : Union[Tuple, list] = tuple([[10, 13, 16, 30, 33, 23],
                                          [30, 61, 62, 45, 59, 119],
                                          [116, 90, 156, 198, 373, 326]])
    # Strides of feature map.
    strides : Union[Tuple, list] = (8, 16, 32)
    # Input data shape for evaluation, use 320, 416, 608...
    # Training is with random shapes from (320 to 608).
    data_shape : int = 416
    # Use synchronize BN across devices.
    syncbn : bool = False
    # Use fixed size(data-shape) throughout the training, which will be faster
    # and require less memory. However, final model will be slightly worse.
    no_random_shape : bool = False
    # Use MXNet AMP for mixed precision training.
    amp : bool = False
    # Whether to enable custom model.
    # custom_model = True
    # whether apply transfer learning from pre-trained models, if True, override other net structures
    transfer : Union[str, None] = 'yolo3_darknet53_coco'
    # NMS settings
    nms_thresh : Union[float, int] = 0.45
    nms_topk : int = 400


@dataclass
class TrainCfg:
    # Training mini-batch size
    batch_size : int = 16
    # Training epochs.
    epochs : int = 20
    # Starting epoch for resuming, default is 0 for new training.
    # You can specify it to 100 for example to start from 100 epoch.
    start_epoch : int = 0
    # Learning rate.
    lr : float = 0.001
    # learning rate scheduler mode. options are step, poly and cosine.
    lr_mode : str = 'step'
    # decay rate of learning rate.
    lr_decay : float = 0.1
    # interval for periodic learning rate decays.
    lr_decay_period : int = 0
    # epochs at which learning rate decays.
    lr_decay_epoch : Union[Tuple, list] = (160, 180)
    # starting warmup learning rate.
    warmup_lr : float = 0.0
    # number of warmup epochs.
    warmup_epochs : int = 0
    # SGD momentum.
    momentum : float = 0.9
    # Weight decay.
    wd : float = 0.0005
    # Logging mini-batch interval.
    log_interval : int = 100
    # Random seed to be fixed.
    seed : int = 233
    # Training images. Use -1 to automatically get the number.
    num_samples : int = -1
    # whether to remove weight decay on bias, and beta/gamma for batchnorm layers.
    no_wd : bool = False
    # whether to enable mixup.
    mixup : bool = False
    # Disable mixup training if enabled in the last N epochs.
    no_mixup_epochs : int = 20
    # Use label smoothing.
    label_smooth : bool = False
    early_stop_patience : int = -1  # epochs with no improvement after which train is early stopped, negative: disabled
    early_stop_min_delta : float = 0.001  # ignore changes less than min_delta for metrics
    # the baseline value for metric, training won't stop if not reaching baseline
    early_stop_baseline : Union[float, int] = 0.0
    early_stop_max_value : Union[float, int] = 1.0  # early stop if reaching max value instantly


@dataclass
class ValidCfg:
    # Batch size during training
    batch_size : int = 16
    # Epoch interval for validation, increase the number
    # will reduce the training time if validation is slow.
    val_interval : int = 1
    # metric, 'voc', 'voc07'
    metric : str = 'voc07'
    # iou_thresh for VOC type metrics
    iou_thresh : float = 0.5


@dataclass
class YOLOv3Cfg:
    yolo3 : YOLOv3 = field(default_factory=YOLOv3)
    train : TrainCfg = field(default_factory=TrainCfg)
    valid : ValidCfg = field(default_factory=ValidCfg)
    # Training dataset. eg. 'coco', 'voc', 'voc_tiny'
    dataset : str = 'voc_tiny'
    # Path of the directory where the dataset is located.
    dataset_root : str = '~/.mxnet/datasets/'
    # Number of data workers, you can use larger number to
    # accelerate data loading, if you CPU and GPUs are powerful.
    num_workers : int = 4
    # Training with GPUs, you can specify (1,3) for example.
    gpus : Union[Tuple, list] = (0, 1, 2, 3)
    # Resume from previously saved parameters if not None.
    # For example, you can resume from ./yolo3_xxx_0123.params
    resume : str = ''
    # Saving parameter prefix.
    save_prefix : str = ''
    # Saving parameters epoch interval, best model will always be saved.
    save_interval : int = 10
    # Use MXNet Horovod for distributed training. Must be run with OpenMPI.
    horovod : bool = False



"""Utils for auto Mask-RCNN estimator"""
# pylint: disable=consider-using-enumerate
import os
import logging

import mxnet as mx
from mxnet import gluon

from .... import data as gdata
from ....data import batchify
from ....data.sampler import SplitSortedBucketSampler
from ....utils.metrics.coco_instance import COCOInstanceMetric

try:
    import horovod.mxnet as hvd
except ImportError:
    hvd = None

try:
    from mpi4py import MPI
except ImportError:
    logging.info('mpi4py is not installed. Use "pip install --no-cache mpi4py" to install')
    MPI = None


def _get_dataset(dataset, args):
    if dataset.lower() == 'coco':
        train_dataset = gdata.COCOInstance(splits='instances_train2017')
        val_dataset = gdata.COCOInstance(splits='instances_val2017', skip_empty=False)
        starting_id = 0
        if args.horovod and MPI:
            length = len(val_dataset)
            shard_len = length // hvd.size()
            rest = length % hvd.size()
            # Compute the start index for this partition
            starting_id = shard_len * hvd.rank() + min(hvd.rank(), rest)
        val_metric = COCOInstanceMetric(val_dataset, os.path.join(args.logdir, args.save_prefix + '_eval'),
                                        use_ext=args.use_ext, starting_id=starting_id)
    else:
        raise NotImplementedError('Dataset: {} not implemented.'.format(dataset))
    if args.horovod and MPI:
        val_dataset = val_dataset.shard(hvd.size(), hvd.rank())
    return train_dataset, val_dataset, val_metric


def _get_dataloader(net, train_dataset, val_dataset, train_transform, val_transform, batch_size,
                    num_shards_per_process, args):
    """Get dataloader."""
    train_bfn = batchify.MaskRCNNTrainBatchify(net, num_shards_per_process)
    train_sampler = \
        SplitSortedBucketSampler(train_dataset.get_im_aspect_ratio(),
                                 batch_size,
                                 num_parts=hvd.size() if args.horovod else 1,
                                 part_index=hvd.rank() if args.horovod else 0,
                                 shuffle=True)
    train_loader = gluon.data.DataLoader(
        train_dataset.transform(train_transform(net.short, net.max_size,
                                                net, ashape=net.ashape, multi_stage=args.mask_rcnn.use_fpn)),
        batch_sampler=train_sampler, batchify_fn=train_bfn, num_workers=args.num_workers)
    val_bfn = batchify.Tuple(*[batchify.Append() for _ in range(2)])
    short = net.short[-1] if isinstance(net.short, (tuple, list)) else net.short
    # validation use 1 sample per device
    val_loader = gluon.data.DataLoader(
        val_dataset.transform(val_transform(short, net.max_size)), num_shards_per_process, False,
        batchify_fn=val_bfn, last_batch='keep', num_workers=args.num_workers)
    return train_loader, val_loader


def _save_params(net, logger, best_map, current_map, epoch, save_interval, prefix):
    current_map = float(current_map)
    if current_map > best_map[0]:
        logger.info('[Epoch {}] mAP {} higher than current best {} saving to {}'.format(
            epoch, current_map, best_map, '{:s}_best.params'.format(prefix)))
        best_map[0] = current_map
        net.save_parameters('{:s}_best.params'.format(prefix))
        with open(prefix + '_best_map.log', 'a') as log_file:
            log_file.write('\n{:04d}:\t{:.4f}'.format(epoch, current_map))
    if save_interval and (epoch + 1) % save_interval == 0:
        logger.info('[Epoch {}] Saving parameters to {}'.format(
            epoch, '{:s}_{:04d}_{:.4f}.params'.format(prefix, epoch, current_map)))
        net.save_parameters('{:s}_{:04d}_{:.4f}.params'.format(prefix, epoch, current_map))


def _stage_data(i, data, ctx_list, pinned_data_stage):
    def _get_chunk(data, storage):
        s = storage.reshape(shape=(storage.size,))
        s = s[:data.size]
        s = s.reshape(shape=data.shape)
        data.copyto(s)
        return s

    if ctx_list[0].device_type == "cpu":
        return data
    if i not in pinned_data_stage:
        pinned_data_stage[i] = [d.as_in_context(mx.cpu_pinned()) for d in data]
        return pinned_data_stage[i]

    storage = pinned_data_stage[i]

    for j in range(len(storage)):
        if data[j].size > storage[j].size:
            storage[j] = data[j].as_in_context(mx.cpu_pinned())

    return [_get_chunk(d, s) for d, s in zip(data, storage)]


def _split_and_load(batch, ctx_list):
    """Split data to 1 batch each device."""
    new_batch = []
    for _, data in enumerate(batch):
        if isinstance(data, (list, tuple)):
            new_data = [x.as_in_context(ctx) for x, ctx in zip(data, ctx_list)]
        else:
            new_data = [data.as_in_context(ctx_list[0])]
        new_batch.append(new_data)
    return new_batch


def _get_lr_at_iter(alpha, lr_warmup_factor=1. / 3.):
    return lr_warmup_factor * (1 - alpha) + alpha



"""
Loss functions used for training COOT model
"""

from torch import nn
import torch

__all__ = ['MaxMarginRankingLoss', 'CycleConsistencyCootLoss', 'IOULoss']


class MaxMarginRankingLoss(nn.Module):
    """MaxMarginRanking loss used in COOT paper (Eq. 1): https://arxiv.org/abs/2011.00597
    Inputs:
        - x_embedding: the embedding tensor for a modality with shape (Batch x feature_dimension).
        - y_embedding: the embedding tensor for another input modality with same shape of x_embedding.
          i.e. in COOT paper x_embedding could be video features and y_embedding language features.
        - margin: constant margin in eq. 1 of COOT paper.
    Outputs:
        - **loss**: loss tensor with shape (batch_size,).
    """
    def __init__(self, use_cuda: bool, margin: float = 0.2):
        super().__init__()
        self.margin = margin
        self.sim = self._cosine_sim
        self.use_cuda = use_cuda

    def _cosine_sim(self, embedding_x1, embedding_x2):
        """ calculate cosine similarity"""
        return embedding_x1.mm(embedding_x2.t())

    def forward(self, x_embedding, y_embedding):
        """ calculate marxmargin ranking loss"""

        scores = self.sim(x_embedding,
                          y_embedding)  # calculate the cosine similarity
        diagonal = scores.diag().view(
            x_embedding.size(0),
            1)  # extract the diagonal scores ( i.e. video_i, sentence_i)
        d1 = diagonal.expand_as(scores)
        d2 = diagonal.t().expand_as(scores)
        cost_x1 = (self.margin + scores - d1).clamp(
            min=0)  # First term in eq.1 of COOT paper
        cost_x2 = (self.margin + scores - d2).clamp(
            min=0)  # Second term in eq.1 of COOT paper
        mask = torch.eye(scores.size(0)) > .5

        if self.use_cuda:
            mask = mask.cuda()

        cost_x1 = cost_x1.masked_fill_(mask, 0)
        cost_x2 = cost_x2.masked_fill_(mask, 0)
        return (cost_x1.sum() + cost_x2.sum()).div(x_embedding.shape[0] *
                                                   y_embedding.shape[0])


class CycleConsistencyCootLoss(nn.Module):
    """Compute CycleConsistency loss in COOT paper (Eq. 8)

    Args:
        clip_emb: Embeddings of clips
        clip_mask: Mask
        clip_lens: Number of clips per each video
        sentence_emb: Embeddings of sentences
        sentence_mask: Mask for input language
        sentence_lens: Number of sentences per paragraph

    Outputs:
        - ** clip to clip loss**: loss tensor with shape (batch_size,).
        - ** sentence to sentence loss**: loss tensor with shape (batch_size,).

    """
    def __init__(self, num_samples=-1, use_cuda=True):
        super().__init__()
        self.num_samples = num_samples
        self.use_cuda = use_cuda
        self.num_samples_tensor = (torch.ones(1) * self.num_samples)
        if self.use_cuda:
            self.num_samples_tensor = self.num_samples_tensor.cuda(
                non_blocking=True)
        self.loss_distance_fn = self._compute_mean_distance_l2
        self.proximity_fn = self._compute_mean_distance_negative_l2
        self.proximity_mask_val = -1e18
        self.softmax = nn.Softmax(dim=-1)

    def _compute_mean_distance_l2(self, c, s):
        return torch.mean((c - s)**2, dim=-1)

    def _compute_mean_distance_negative_l2(self, c, s):
        return -self._compute_mean_distance_l2(c, s)

    def forward(self, clip_emb, clip_mask, clip_lens, sentence_emb,
                sentence_mask, sentence_lens):
        """ calculate the loss"""

        clip_max_len, _ = torch.max(clip_lens, dim=-1)
        sentence_max_len, _ = torch.max(sentence_lens, dim=-1)
        clip_sentence_nn, _, _ = self._get_soft_nn(clip_emb, clip_mask,
                                                   sentence_emb, sentence_mask)
        _, clip_beta, _ = self._get_soft_nn(clip_sentence_nn, clip_mask,
                                            clip_emb, clip_mask)
        clip_clip_loss = self._get_loss(clip_mask, clip_lens, clip_max_len,
                                        clip_beta)
        sentence_clip_nn, _, _ = self._get_soft_nn(sentence_emb, sentence_mask,
                                                   clip_emb, clip_mask)
        _, sentence_beta, _ = self._get_soft_nn(sentence_clip_nn,
                                                sentence_mask, sentence_emb,
                                                sentence_mask)
        sentence_sentence_loss = self._get_loss(sentence_mask, sentence_lens,
                                                sentence_max_len,
                                                sentence_beta)
        return clip_clip_loss, sentence_sentence_loss

    def _get_mxn_repr(self, source_emb, source_mask, target_emb, target_mask):
        source_rep = source_emb.unsqueeze(2)
        target_rep = target_emb.unsqueeze(1)
        total_mask = source_mask.unsqueeze(2).bool() & target_mask.unsqueeze(
            1).bool()
        return source_rep, target_rep, total_mask

    def _get_soft_nn(self, source_emb, source_mask, target_emb, target_mask):
        source_rep, target_rep, total_mask = self._get_mxn_repr(
            source_emb, source_mask, target_emb, target_mask)
        distance = self.proximity_fn(source_rep, target_rep)
        distance.masked_fill_(total_mask == 0, self.proximity_mask_val)
        weights_alpha = self.softmax(distance)
        soft_nn = target_emb.unsqueeze(dim=1) * weights_alpha.unsqueeze(dim=3)
        soft_nn = torch.sum(soft_nn, dim=2)
        return soft_nn, weights_alpha, distance

    def _get_loss(self, emb_mask, emb_lens, emb_max_len, beta):
        idx_orig = torch.arange(emb_max_len)
        batch_size, _ = emb_mask.shape
        if self.use_cuda:
            idx_orig = idx_orig.cuda(non_blocking=True)
        idx_orig.unsqueeze_(0)
        index_nn = torch.sum(idx_orig.unsqueeze(1) * beta, dim=-1)
        idx_nn_rep, idx_orig_rep, emb_mask_rep = self._get_mxn_repr(
            index_nn, emb_mask, idx_orig, emb_mask)
        distance = self.loss_distance_fn(idx_nn_rep.unsqueeze(-1),
                                         idx_orig_rep.unsqueeze(-1))
        distance.masked_fill_(emb_mask_rep == 0, 0)
        l_seq = distance.diagonal(dim1=-2, dim2=-1)
        if self.num_samples != -1:
            n_samp = torch.min(emb_lens, self.num_samples_tensor)
            total_loss = 0
            for _, (c_loss, c_mask,
                    c_nsamp) in enumerate(zip(l_seq, emb_mask, n_samp)):
                idx = torch.multinomial(c_mask, int(c_nsamp))
                total_loss += c_loss[idx].mean()
            return total_loss / batch_size

        else:
            return (l_seq.sum(dim=-1) / emb_lens).mean(dim=-1)


class IOULoss(nn.Module):
    """
    Intersetion Over Union (IoU) loss which supports three
    different IoU computations:

    * IoU
    * Linear IoU
    * gIoU
    """
    def __init__(self, loc_loss_type='iou'):
        super(IOULoss, self).__init__()
        self.loc_loss_type = loc_loss_type

    def forward(self, pred, target, weight=None):
        """
        Args:
            pred: Nx4 predicted bounding boxes
            target: Nx4 target bounding boxes
            weight: N loss weight for each instance
        """
        pred_left = pred[:, 0]
        pred_top = pred[:, 1]
        pred_right = pred[:, 2]
        pred_bottom = pred[:, 3]

        target_left = target[:, 0]
        target_top = target[:, 1]
        target_right = target[:, 2]
        target_bottom = target[:, 3]

        target_aera = (target_left + target_right) * \
                      (target_top + target_bottom)
        pred_aera = (pred_left + pred_right) * \
                    (pred_top + pred_bottom)

        w_intersect = torch.min(pred_left, target_left) + \
                      torch.min(pred_right, target_right)
        h_intersect = torch.min(pred_bottom, target_bottom) + \
                      torch.min(pred_top, target_top)

        g_w_intersect = torch.max(pred_left, target_left) + \
                        torch.max(pred_right, target_right)
        g_h_intersect = torch.max(pred_bottom, target_bottom) + \
                        torch.max(pred_top, target_top)
        ac_uion = g_w_intersect * g_h_intersect

        area_intersect = w_intersect * h_intersect
        area_union = target_aera + pred_aera - area_intersect

        ious = (area_intersect + 1.0) / (area_union + 1.0)
        gious = ious - (ac_uion - area_union) / ac_uion
        if self.loc_loss_type == 'iou':
            losses = -torch.log(ious)
        elif self.loc_loss_type == 'linear_iou':
            losses = 1 - ious
        elif self.loc_loss_type == 'giou':
            losses = 1 - gious
        else:
            raise NotImplementedError

        if weight is not None:
            return (losses * weight).sum()
        else:
            return losses.sum()



"""
Provide a general object detector interface for SMOT
"""
# pylint: disable=unused-wildcard-import,wildcard-import
import logging
import numpy as np

import mxnet as mx

from gluoncv.data import COCODetection
from .utils import mxnet_frame_preprocessing, timeit_context
from .utils import remap_bboxes as _remap_bboxes
from .presets import *


ssd_base_models = {'ssd_300_vgg16_atrous_voc': ssd_300_vgg16_atrous_voc,
                   'ssd_300_vgg16_atrous_coco': ssd_300_vgg16_atrous_coco,
                   'ssd_300_vgg16_atrous_custom': ssd_300_vgg16_atrous_custom,
                   'ssd_512_vgg16_atrous_voc': ssd_512_vgg16_atrous_voc,
                   'ssd_512_vgg16_atrous_coco': ssd_512_vgg16_atrous_coco,
                   'ssd_512_vgg16_atrous_custom': ssd_512_vgg16_atrous_custom,
                   'ssd_512_resnet18_v1_voc': ssd_512_resnet18_v1_voc,
                   'ssd_512_resnet18_v1_coco': ssd_512_resnet18_v1_coco,
                   'ssd_512_resnet50_v1_voc': ssd_512_resnet50_v1_voc,
                   'ssd_512_resnet50_v1_coco': ssd_512_resnet50_v1_coco,
                   'ssd_512_resnet50_v1_custom': ssd_512_resnet50_v1_custom,
                   'ssd_512_resnet101_v2_voc': ssd_512_resnet101_v2_voc,
                   'ssd_512_resnet152_v2_voc': ssd_512_resnet152_v2_voc,
                   'ssd_512_mobilenet1.0_voc': ssd_512_mobilenet1_0_voc,
                   'ssd_512_mobilenet1.0_coco': ssd_512_mobilenet1_0_coco,
                   'ssd_300_mobilenet1.0_lite_coco': ssd_300_mobilenet1_0_lite_coco,
                   'ssd_512_mobilenet1.0_custom': ssd_512_mobilenet1_0_custom,
                   'ssd_300_mobilenet0.25_voc': ssd_300_mobilenet0_25_voc,
                   'ssd_300_mobilenet0.25_coco': ssd_300_mobilenet0_25_coco,
                   'ssd_300_mobilenet0.25_custom': ssd_300_mobilenet0_25_custom,
                   'ssd_300_resnet34_v1b_voc': ssd_300_resnet34_v1b_voc,
                   'ssd_300_resnet34_v1b_coco': ssd_300_resnet34_v1b_coco,
                   'ssd_300_resnet34_v1b_custom': ssd_300_resnet34_v1b_custom,}


# pylint: disable=line-too-long,missing-class-docstring,missing-module-docstring,missing-function-docstring,unused-argument
def get_net(classes, model_name="", use_pretrained=False, param_path="",
            ctx=None, **kwargs):
    assert model_name in ssd_base_models, "the model name is not supported, where the supported models are {}".format(ssd_base_models.keys())
    if use_pretrained:
        # use off-the-shelf GluonCV pretrained SSD models
        net = ssd_base_models[model_name](pretrained=use_pretrained,
                                          pretrained_base=False, ctx=ctx, **kwargs)
    else:
        # use finetuned model weights or customized trained model weights
        net = ssd_base_models[model_name](pretrained_base=False, ctx=ctx, **kwargs)
        assert param_path != '', "Please provide the pretrained model weights if you are not using GluonCV pretrained detectors."
        net.load_parameters(param_path, ctx=ctx)
    net.hybridize()
    return net


def _remap_keypoints(keypoints, padded_w, padded_h, expand, data_shape, ratio):
    """
    Remap bboxes in (x0, y0, x1, y1) format into the input image space
    Parameters
    ----------
    bboxes
    padded_w
    padded_h
    expand

    Returns
    -------

    """
    keypoints[:, 0::2] *= padded_w / (data_shape * ratio)
    keypoints[:, 1::2] *= padded_h / data_shape
    keypoints[:, 0::2] -= expand[0]
    keypoints[:, 1::2] -= expand[1]

    return keypoints


class GeneralDetector:

    def __init__(self, gpu_id,
                 aspect_ratio=1.,
                 data_shape=512,
                 model_name="",
                 use_pretrained=False,
                 param_path=""):
        self.ctx = mx.gpu(gpu_id)

        self.net = get_net(classes=COCODetection.CLASSES,
                           ctx=self.ctx,
                           model_name=model_name,
                           use_pretrained=use_pretrained,
                           param_path=param_path)

        self.anchor_tensor = None
        self._anchor_image_shape = (1, 1)
        self._anchor_num = 1

        self.mean_mx = mx.nd.array(np.array([0.485, 0.456, 0.406])).as_in_context(self.ctx)
        self.std_mx = mx.nd.array(np.array([0.229, 0.224, 0.225])).as_in_context(self.ctx)
        self.ratio = aspect_ratio
        self.data_shape = data_shape

    def run_detection(self, image, tracking_box_indices, tracking_box_weights, tracking_box_classes):
        """

        Parameters
        ----------
        image: RGB images

        Returns
        -------

        """
        # pylint: disable=logging-format-interpolation
        with timeit_context("preprocess"):
            data_tensor, padded_w, padded_h, expand = mxnet_frame_preprocessing(image, self.data_shape, self.ratio,
                                                                                self.mean_mx, self.std_mx, self.ctx)

            logging.info("input tensor shape {}".format(data_tensor.shape))
            mx.nd.waitall()

        with timeit_context("network"):
            real_tracking_indices = tracking_box_indices + tracking_box_classes * self._anchor_num
            ids, scores, detection_bboxes, detection_anchor_indices, tracking_results, anchors = self.net(
                data_tensor.as_in_context(self.ctx), real_tracking_indices, tracking_box_weights)

            tracking_bboxes = tracking_results[:, [2, 3, 4, 5, 1]]

            detection_bboxes = _remap_bboxes(detection_bboxes[0, :, :],
                                             padded_w, padded_h, expand,
                                             self.data_shape, self.ratio)
            tracking_bboxes = _remap_bboxes(tracking_bboxes,
                                            padded_w, padded_h, expand,
                                            self.data_shape, self.ratio)
            mx.nd.waitall()

        # set anchors if needed
        if self._anchor_image_shape != (image.shape[:2]):
            self._anchor_image_shape = image.shape[:2]
            # initialize the anchor tensor for assignment
            self.anchor_tensor = anchors[0, :, :]
            half_w = self.anchor_tensor[:, 2] / 2
            half_h = self.anchor_tensor[:, 3] / 2
            center_x = self.anchor_tensor[:, 0].copy()
            center_y = self.anchor_tensor[:, 1].copy()

            # anchors are in the original format of (center_x, center_y, w, h)
            # translate them to (x0, y0, x1, y1)
            self.anchor_tensor[:, 0] = center_x - half_w
            self.anchor_tensor[:, 1] = center_y - half_h

            self.anchor_tensor[:, 2] = center_x + half_w
            self.anchor_tensor[:, 3] = center_y + half_h

            self.anchor_tensor = _remap_bboxes(self.anchor_tensor, padded_w, padded_h, expand,
                                               self.data_shape, self.ratio)
            self._anchor_num = self.anchor_tensor.shape[0]

        return ids[0], scores[0], detection_bboxes, tracking_bboxes, detection_anchor_indices[0].asnumpy()



"""Registry of most widely used datasets"""
import os
import shutil
import tarfile
import zipfile
from ...data import VOCDetection, COCODetection
from ...utils import download as _download, makedirs

__all__ = ['get_dataset', 'list_dataset']

_DATASETS = {}


def list_dataset():
    """List available dataset registered in auto data.

    Returns
    -------
    list
        List of names.

    """
    return list(_DATASETS.keys())

def get_dataset(ds_name, train=True, download=True):
    """Get the auto dataset with `ds_name`.

    Parameters
    ----------
    ds_name : str
        The name of dataset, you may view registered with `list_dataset`.
    train : bool, default is `True`
        If `True`, get the training set, if `False`, try to get the validation set.
    download : bool, default is `True`
        If `True`, download the dataset automatically if not found on disk.

    """
    if not ds_name in _DATASETS:
        choices = str(_DATASETS.keys())
        raise ValueError('{} not found. Available choices: {}'.format(ds_name, choices))
    ret = _DATASETS[ds_name](download=download)
    if not isinstance(ret, (list, tuple)):
        if train:
            return ret
        else:
            raise ValueError('No validation dataset available for {}'.format(ds_name))
    if train:
        return ret[0]
    else:
        if len(ret) < 2:
            raise ValueError('No validation dataset available for {}'.format(ds_name))
        return ret[1]

def _register_dataset(ds_name, fn):
    if ds_name in _DATASETS:
        raise ValueError('{} already been registered'.format(ds_name))
    assert callable(fn), "register with non-function is not allowed"
    _DATASETS[ds_name] = fn

def _pascal_0712_detection(download=True):
    root = os.environ.get('MXNET_HOME', os.path.join('~', '.mxnet'))
    root = os.path.expanduser(os.path.join(root, 'datasets', 'voc'))
    if not os.path.isdir(root):
        if not download:
            raise OSError('No `pascal` dataset found on disk and `download=False`')
        # download automatically
        makedirs(root)
        filename = _download('https://s3.amazonaws.com/fast-ai-imagelocal/pascal-voc.tgz',
                             path=root, overwrite=True)
        with tarfile.open(filename) as tar:
            tar.extractall(path=root)
        shutil.move(os.path.join(root, 'pascal-voc', 'VOC2007'), os.path.join(root, 'VOC2007'))
        shutil.move(os.path.join(root, 'pascal-voc', 'VOC2012'), os.path.join(root, 'VOC2012'))
        shutil.rmtree(os.path.join(root, 'pascal-voc'))
    return VOCDetection(root=root, splits=[(2007, 'trainval'), (2012, 'trainval')]), \
           VOCDetection(root=root, splits=[(2007, 'test')])

def _coco_2017_detection(download=True):
    root = os.environ.get('MXNET_HOME', os.path.join('~', '.mxnet'))
    root = os.path.expanduser(os.path.join(root, 'datasets', 'coco'))
    if not os.path.isdir(root):
        if not download:
            raise OSError('No `coco` dataset found on disk and `download=False`')
        # download automatically
        makedirs(root)
        _DOWNLOAD_URLS = [
            ('http://images.cocodataset.org/zips/train2017.zip',
             '10ad623668ab00c62c096f0ed636d6aff41faca5'),
            ('http://images.cocodataset.org/annotations/annotations_trainval2017.zip',
             '8551ee4bb5860311e79dace7e79cb91e432e78b3'),
            ('http://images.cocodataset.org/zips/val2017.zip',
             '4950dc9d00dbe1c933ee0170f5797584351d2a41'),
        ]
        for url, checksum in _DOWNLOAD_URLS:
            filename = _download(url, path=root, overwrite=True, sha1_hash=checksum)
            # extract
            with zipfile.ZipFile(filename) as zf:
                zf.extractall(path=root)
    return COCODetection(root=root, splits='instances_train2017'), \
           COCODetection(root=root, splits='instances_val2017', skip_empty=False)

_register_dataset('pascal_voc_0712_detection', _pascal_0712_detection)
_register_dataset('coco_2017_detection', _coco_2017_detection)



from __future__ import print_function
from __future__ import division

import os
import mxnet as mx
import numpy as np

import gluoncv as gcv
from gluoncv import data
from .tiny_datasets import COCODetectionTiny, COCOInstanceTiny, VOCDetectionTiny, VOCSegmentationTiny
import os.path as osp
import unittest


def test_pascal_voc_detection():
    train = VOCDetectionTiny()
    name = str(train)
    val = VOCDetectionTiny(splits=(('tiny_motorbike', 'test'), ))
    name = str(val)

    assert train.classes == val.classes

    for _ in range(10):
        index = np.random.randint(0, len(train))
        _ = train[index]

    for _ in range(10):
        index = np.random.randint(0, len(val))
        _ = val[index]


def test_coco_detection():
    # use valid only, loading training split is very slow
    val = COCODetectionTiny()
    name = str(val)
    assert len(val.classes) > 0

    for _ in range(10):
        index = np.random.randint(0, len(val))
        _ = val[index]

def test_coco_instance():
    # use valid only, loading training split is very slow
    val = COCOInstanceTiny()
    name = str(val)
    assert len(val.classes) > 0

    for _ in range(10):
        index = np.random.randint(0, len(val))
        _ = val[index]

@unittest.skip("temporarily disabled due to drawing compatibility")
def test_voc_segmentation():

    # use valid only, loading training split is very slow
    val = VOCSegmentationTiny()
    name = str(val)
    assert len(val.classes) > 0

    for _ in range(10):
        index = np.random.randint(0, len(val))
        _ = val[index]

def test_voc_aug_segmentation():
    pass
    # if not osp.isdir(osp.expanduser('~/.mxnet/datasets/voc')):
    #     return
    #
    # # use valid only, loading training split is very slow
    # val = data.VOCAugSegmentation(split='train')
    # name = str(val)
    # assert len(val.classes) > 0
    #
    # for _ in range(10):
    #     index = np.random.randint(0, len(val))
    #     _ = val[index]

def test_ade_segmentation():
    if not osp.isdir(osp.expanduser('~/.mxnet/datasets/ade')):
        return

    # use valid only, loading training split is very slow
    val = data.ADE20KSegmentation(split='train')
    name = str(val)

    for _ in range(10):
        index = np.random.randint(0, len(val))
        _ = val[index]

def test_citys_segmentation():
    if not osp.isdir(osp.expanduser('~/.mxnet/datasets/ade')):
        return

    # use valid only, loading training split is very slow
    val = data.CitySegmentation(split='train')
    name = str(val)

    for _ in range(10):
        index = np.random.randint(0, len(val))
        _ = val[index]

def test_lst_detection():
    dog_label = [130, 220, 320, 530]
    bike_label = [115, 120, 580, 420]
    car_label = [480, 80, 700, 170]
    all_boxes = np.array([dog_label, bike_label, car_label])
    all_ids = np.array([0, 1, 2])
    class_names = ['dog', 'bike', 'car']
    im_fname = gcv.utils.download('https://github.com/dmlc/web-data/blob/master/' +
                              'gluoncv/datasets/dog.jpg?raw=true',
                              path='dog.jpg')
    img = mx.image.imread(im_fname)

    def write_line(img_path, im_shape, boxes, ids, idx):
        h, w, c = im_shape
        # for header, we use minimal length 2, plus width and height
        # with A: 4, B: 5, C: width, D: height
        A = 4
        B = 5
        C = w
        D = h
        # concat id and bboxes
        labels = np.hstack((ids.reshape(-1, 1), boxes)).astype('float')
        # normalized bboxes (recommended)
        labels[:, (1, 3)] /= float(w)
        labels[:, (2, 4)] /= float(h)
        # flatten
        labels = labels.flatten().tolist()
        str_idx = [str(idx)]
        str_header = [str(x) for x in [A, B, C, D]]
        str_labels = [str(x) for x in labels]
        str_path = [img_path]
        line = '\t'.join(str_idx + str_header + str_labels + str_path) + '\n'
        return line

    with open('val.lst', 'w') as fw:
        for i in range(4):
            line = write_line('dog.jpg', img.shape, all_boxes, all_ids, i)
            print(line)
            fw.write(line)
    from gluoncv.data import LstDetection
    lst_dataset = LstDetection('val.lst', root=os.path.expanduser('.'))
    assert len(lst_dataset) == 4
    lst_dataset.transform(lambda x, y : (x, y))
    try:
        os.remove('dog.jpg')
        os.remove('val.lst')
    except IOError:
        pass

if __name__ == '__main__':
    import nose
    nose.runmodule()



# pylint: disable=line-too-long
"""
Utility functions for video action recognition task
"""
import os
import logging
import time
import numpy as np

import torch

from ..utils import AverageMeter, accuracy

logger = logging.getLogger(__name__)


def train_classification(base_iter,
                         model,
                         dataloader,
                         epoch,
                         criterion,
                         optimizer,
                         cfg,
                         writer=None):
    """Task of training video classification"""
    batch_time = AverageMeter()
    data_time = AverageMeter()
    losses = AverageMeter()
    top1 = AverageMeter()
    top5 = AverageMeter()

    model.train()
    end = time.time()
    for step, data in enumerate(dataloader):
        base_iter = base_iter + 1

        train_batch = data[0].cuda()
        train_label = data[1].cuda()
        data_time.update(time.time() - end)

        outputs = model(train_batch)
        loss = criterion(outputs, train_label)
        if cfg.CONFIG.DATA.NUM_CLASSES < 5:
            prec1, prec5 = accuracy(outputs.data, train_label, topk=(1, 1))
            # Tricky solution for datasets with less than 5 classes, top5 acc is always set to 100%
            prec5 = 100
        else:
            prec1, prec5 = accuracy(outputs.data, train_label, topk=(1, 5))

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        losses.update(loss.item(), train_label.size(0))
        top1.update(prec1.item(), train_label.size(0))
        top5.update(prec5.item(), train_label.size(0))

        batch_time.update(time.time() - end)
        end = time.time()
        if step % cfg.CONFIG.LOG.DISPLAY_FREQ == 0 and cfg.DDP_CONFIG.GPU_WORLD_RANK == 0:
            logger.info('-------------------------------------------------------')
            for param in optimizer.param_groups:
                lr = param['lr']
            print_string = 'lr: {lr:.5f}'.format(lr=lr)
            logger.info(print_string)
            print_string = 'Epoch: [{0}][{1}/{2}]'.format(
                epoch, step + 1, len(dataloader))
            logger.info(print_string)
            print_string = 'data_time: {data_time:.3f}, batch time: {batch_time:.3f}'.format(
                data_time=data_time.val, batch_time=batch_time.val)
            logger.info(print_string)
            print_string = 'loss: {loss:.5f}'.format(loss=losses.avg)
            logger.info(print_string)
            print_string = 'Top-1 accuracy: {top1_acc:.2f}%, Top-5 accuracy: {top5_acc:.2f}%'.format(
                top1_acc=top1.avg, top5_acc=top5.avg)
            logger.info(print_string)
            iteration = base_iter
            writer.add_scalar('train_loss_iteration', losses.avg, iteration)
            writer.add_scalar('train_top1_acc_iteration', top1.avg, iteration)
            writer.add_scalar('train_top5_acc_iteration', top5.avg, iteration)
            writer.add_scalar('train_batch_size_iteration',
                              train_label.size(0), iteration)
            writer.add_scalar('learning_rate', lr, iteration)
    return base_iter


def validation_classification(model, val_dataloader, epoch, criterion, cfg,
                              writer):
    """Task of validating video classification"""
    batch_time = AverageMeter()
    data_time = AverageMeter()
    losses = AverageMeter()
    top1 = AverageMeter()
    top5 = AverageMeter()
    model.eval()

    end = time.time()
    with torch.no_grad():
        for step, data in enumerate(val_dataloader):
            data_time.update(time.time() - end)
            val_batch = data[0].cuda()
            val_label = data[1].cuda()
            outputs = model(val_batch)

            loss = criterion(outputs, val_label)
            if cfg.CONFIG.DATA.NUM_CLASSES < 5:
                prec1a, prec5a = accuracy(outputs.data, val_label, topk=(1, 1))
                # Tricky solution for datasets with less than 5 classes, top5 acc is always set to 100%
                prec5a = 100
            else:
                prec1a, prec5a = accuracy(outputs.data, val_label, topk=(1, 5))

            losses.update(loss.item(), val_batch.size(0))
            top1.update(prec1a.item(), val_batch.size(0))
            top5.update(prec5a.item(), val_batch.size(0))
            batch_time.update(time.time() - end)
            end = time.time()

            if step % cfg.CONFIG.LOG.DISPLAY_FREQ == 0 and cfg.DDP_CONFIG.GPU_WORLD_RANK == 0:
                logger.info('----validation----')
                print_string = 'Epoch: [{0}][{1}/{2}]'.format(
                    epoch, step + 1, len(val_dataloader))
                logger.info(print_string)
                print_string = 'data_time: {data_time:.3f}, batch time: {batch_time:.3f}'.format(
                    data_time=data_time.val, batch_time=batch_time.val)
                logger.info(print_string)
                print_string = 'loss: {loss:.5f}'.format(loss=losses.avg)
                logger.info(print_string)
                print_string = 'Top-1 accuracy: {top1_acc:.2f}%, Top-5 accuracy: {top5_acc:.2f}%'.format(
                    top1_acc=top1.avg, top5_acc=top5.avg)
                logger.info(print_string)

        eval_path = cfg.CONFIG.LOG.EVAL_DIR
        if not os.path.exists(eval_path):
            os.makedirs(eval_path)

        with open(
                os.path.join(eval_path,
                             "{}.txt".format(cfg.DDP_CONFIG.GPU_WORLD_RANK)),
                'w') as f:
            f.write("{} {} {}\n".format(losses.avg, top1.avg, top5.avg))
        torch.distributed.barrier()

        loss_lst, top1_lst, top5_lst = [], [], []
        if cfg.DDP_CONFIG.GPU_WORLD_RANK == 0 and writer is not None:
            print("Collecting validation numbers")
            for x in range(cfg.DDP_CONFIG.GPU_WORLD_SIZE):
                data = open(os.path.join(
                    eval_path,
                    "{}.txt".format(x))).readline().strip().split(" ")
                data = [float(x) for x in data]
                loss_lst.append(data[0])
                top1_lst.append(data[1])
                top5_lst.append(data[2])
            print("Global result:")
            print_string = 'loss: {loss:.5f}'.format(loss=np.mean(loss_lst))
            print(print_string)
            print_string = 'Top-1 accuracy: {top1_acc:.2f}%, Top-5 accuracy: {top5_acc:.2f}%'.format(
                top1_acc=np.mean(top1_lst), top5_acc=np.mean(top5_lst))
            print(print_string)
            writer.add_scalar('val_loss_epoch', np.mean(loss_lst), epoch)
            writer.add_scalar('val_top1_acc_epoch', np.mean(top1_lst), epoch)
            writer.add_scalar('val_top5_acc_epoch', np.mean(top5_lst), epoch)


def test_classification(model, test_loader, criterion, cfg, file):
    """Task of testing video classification"""
    batch_time = AverageMeter()
    data_time = AverageMeter()
    losses = AverageMeter()
    top1 = AverageMeter()
    top5 = AverageMeter()
    model.eval()

    end = time.time()
    final_result = []

    with torch.no_grad():
        for step, (inputs, labels, ids, chunk_nb,
                   split_nb) in enumerate(test_loader):
            data_time.update(time.time() - end)
            val_batch = inputs.cuda()
            val_label = labels.cuda()
            outputs = model(val_batch)
            loss = criterion(outputs, val_label)

            for i in range(outputs.size(0)):
                string = "{} {} {} {} {}\n".format(ids[i], \
                                                   str(outputs.data[i].cpu().numpy().tolist()), \
                                                   str(int(labels[i].cpu().numpy())), \
                                                   str(int(chunk_nb[i].cpu().numpy())), \
                                                   str(int(split_nb[i].cpu().numpy())))
                final_result.append(string)

            prec1, prec5 = accuracy(outputs.data, val_label, topk=(1, 5))
            losses.update(loss.item(), val_batch.size(0))
            top1.update(prec1.item(), val_batch.size(0))
            top5.update(prec5.item(), val_batch.size(0))
            batch_time.update(time.time() - end)
            end = time.time()
            if step % cfg.CONFIG.LOG.DISPLAY_FREQ == 0 and cfg.DDP_CONFIG.GPU_WORLD_RANK == 0:
                print('----Testing----')
                print_string = 'Epoch: [{0}][{1}/{2}]'.format(
                    0, step + 1, len(test_loader))
                print(print_string)
                print_string = 'data_time: {data_time:.3f}, batch time: {batch_time:.3f}'.format(
                    data_time=data_time.val, batch_time=batch_time.val)
                print(print_string)
                print_string = 'loss: {loss:.5f}'.format(loss=losses.avg)
                print(print_string)
                print_string = 'Top-1 accuracy: {top1_acc:.2f}%, Top-5 accuracy: {top5_acc:.2f}%'.format(
                    top1_acc=top1.avg, top5_acc=top5.avg)
                print(print_string)
    if not os.path.exists(file):
        os.mknod(file)
    with open(file, 'w') as f:
        f.write("{}, {}\n".format(top1.avg, top5.avg))
        for line in final_result:
            f.write(line)



# adapted from https://github.com/facebookresearch/detectron2/blob/master/detectron2/data/detection_utils.py
"""
Common data processing utilities that are used in a
typical object detection data pipeline.
"""
# pylint: disable=pointless-string-statement
import os
import logging
import numpy as np
from PIL import Image

import torch

from ..structures import (
    BitMasks,
    Boxes,
    BoxMode,
    Instances,
    Keypoints,
    PolygonMasks,
    polygons_to_bitmask,
)

from ..transforms import instance_transforms as T
from ..registry.catalog import MetadataCatalog

__all__ = [
    "SizeMismatchError",
    "convert_image_to_rgb",
    "check_image_size",
    "transform_proposals",
    "transform_instance_annotations",
    "annotations_to_instances",
    # "annotations_to_instances_rotated",
    "build_augmentation",
    "build_augmentation_ranged_clip",
    "create_keypoint_hflip_indices",
    "filter_empty_instances",
    "read_image",
    "try_import_pycocotools"
]

def try_import(package, message=None):
    """Try import specified package, with custom message support.

    Parameters
    ----------
    package : str
        The name of the targeting package.
    message : str, default is None
        If not None, this function will raise customized error message when import error is found.


    Returns
    -------
    module if found, raise ImportError otherwise

    """
    try:
        return __import__(package)
    except ImportError as e:
        if not message:
            raise e
        raise ImportError(message)

def import_try_install(package, extern_url=None):
    """Try import the specified package.
    If the package not installed, try use pip to install and import if success.

    Parameters
    ----------
    package : str
        The name of the package trying to import.
    extern_url : str or None, optional
        The external url if package is not hosted on PyPI.
        For example, you can install a package using:
         "pip install git+http://github.com/user/repo/tarball/master/egginfo=xxx".
        In this case, you can pass the url to the extern_url.

    Returns
    -------
    <class 'Module'>
        The imported python module.

    """
    import tempfile
    import portalocker
    lockfile = os.path.join(tempfile.gettempdir(), package + '_install.lck')
    with portalocker.Lock(lockfile):
        try:
            return __import__(package)
        except ImportError:
            try:
                from pip import main as pipmain
            except ImportError:
                from pip._internal import main as pipmain
                from types import ModuleType
                # fix for pip 19.3
                if isinstance(pipmain, ModuleType):
                    from pip._internal.main import main as pipmain

            # trying to install package
            url = package if extern_url is None else extern_url
            pipmain(['install', '--user', url])  # will raise SystemExit Error if fails

            # trying to load again
            try:
                return __import__(package)
            except ImportError:
                import sys
                import site
                user_site = site.getusersitepackages()
                if user_site not in sys.path:
                    sys.path.append(user_site)
                return __import__(package)
    return __import__(package)

"""Import helper for pycocotools"""
# NOTE: for developers
# please do not import any pycocotools in __init__ because we are trying to lazy
# import pycocotools to avoid install it for other users who may not use it.
# only import when you actually use it

def try_import_pycocotools():
    """Tricks to optionally install and import pycocotools"""
    # first we can try import pycocotools
    try:
        import pycocotools as _
    except ImportError:
        # we need to install pycootools, which is a bit tricky
        # pycocotools sdist requires Cython, numpy(already met)
        import_try_install('cython')
        # pypi pycocotools is not compatible with windows
        win_url = 'git+https://github.com/zhreshold/cocoapi.git#subdirectory=PythonAPI'
        try:
            if os.name == 'nt':
                import_try_install('pycocotools', win_url)
            else:
                import_try_install('pycocotools')
        except ImportError:
            faq = 'cocoapi FAQ'
            raise ImportError('Cannot import or install pycocotools, please refer to %s.' % faq)


class SizeMismatchError(ValueError):
    """
    When loaded image has difference width/height compared with annotation.
    """


# https://en.wikipedia.org/wiki/YUV#SDTV_with_BT.601
_M_RGB2YUV = [[0.299, 0.587, 0.114], [-0.14713, -0.28886, 0.436], [0.615, -0.51499, -0.10001]]
_M_YUV2RGB = [[1.0, 0.0, 1.13983], [1.0, -0.39465, -0.58060], [1.0, 2.03211, 0.0]]

# https://www.exiv2.org/tags.html
_EXIF_ORIENT = 274  # exif 'Orientation' tag


def convert_PIL_to_numpy(image, format):
    """
    Convert PIL image to numpy array of target format.

    Args:
        image (PIL.Image): a PIL image
        format (str): the format of output image

    Returns:
        (np.ndarray): also see `read_image`
    """
    if format is not None:
        # PIL only supports RGB, so convert to RGB and flip channels over below
        conversion_format = format
        if format in ["BGR", "YUV-BT.601"]:
            conversion_format = "RGB"
        image = image.convert(conversion_format)
    image = np.asarray(image)
    # PIL squeezes out the channel dimension for "L", so make it HWC
    if format == "L":
        image = np.expand_dims(image, -1)

    # handle formats not supported by PIL
    elif format == "BGR":
        # flip channels if needed
        image = image[:, :, ::-1]
    elif format == "YUV-BT.601":
        image = image / 255.0
        image = np.dot(image, np.array(_M_RGB2YUV).T)

    return image


def convert_image_to_rgb(image, format):
    """
    Convert an image from given format to RGB.

    Args:
        image (np.ndarray or Tensor): an HWC image
        format (str): the format of input image, also see `read_image`

    Returns:
        (np.ndarray): (H,W,3) RGB image in 0-255 range, can be either float or uint8
    """
    if isinstance(image, torch.Tensor):
        image = image.cpu().numpy()
    if format == "BGR":
        image = image[:, :, [2, 1, 0]]
    elif format == "YUV-BT.601":
        image = np.dot(image, np.array(_M_YUV2RGB).T)
        image = image * 255.0
    else:
        if format == "L":
            image = image[:, :, 0]
        image = image.astype(np.uint8)
        image = np.asarray(Image.fromarray(image, mode=format).convert("RGB"))
    return image


def _apply_exif_orientation(image):
    """
    Applies the exif orientation correctly.

    This code exists per the bug:
      https://github.com/python-pillow/Pillow/issues/3973
    with the function `ImageOps.exif_transpose`. The Pillow source raises errors with
    various methods, especially `tobytes`

    Function based on:
      https://github.com/wkentaro/labelme/blob/v4.5.4/labelme/utils/image.py#L59
      https://github.com/python-pillow/Pillow/blob/7.1.2/src/PIL/ImageOps.py#L527

    Args:
        image (PIL.Image): a PIL image

    Returns:
        (PIL.Image): the PIL image with exif orientation applied, if applicable
    """
    if not hasattr(image, "getexif"):
        return image

    exif = image.getexif()

    if exif is None:
        return image

    orientation = exif.get(_EXIF_ORIENT)

    method = {
        2: Image.FLIP_LEFT_RIGHT,
        3: Image.ROTATE_180,
        4: Image.FLIP_TOP_BOTTOM,
        5: Image.TRANSPOSE,
        6: Image.ROTATE_270,
        7: Image.TRANSVERSE,
        8: Image.ROTATE_90,
    }.get(orientation)

    if method is not None:
        return image.transpose(method)
    return image


def read_image(file_name, format=None):
    """
    Read an image into the given format.
    Will apply rotation and flipping if the image has such exif information.

    Args:
        file_name (str): image file path
        format (str): one of the supported image modes in PIL, or "BGR" or "YUV-BT.601".

    Returns:
        image (np.ndarray): an HWC image in the given format, which is 0-255, uint8 for
            supported image modes in PIL or "BGR"; float (0-1 for Y) for YUV-BT.601.
    """
    with open(file_name, "rb") as f:
        image = Image.open(f)

        # work around this bug: https://github.com/python-pillow/Pillow/issues/3973
        image = _apply_exif_orientation(image)

        return convert_PIL_to_numpy(image, format)


def check_image_size(dataset_dict, image):
    """
    Raise an error if the image does not match the size specified in the dict.
    """
    if "width" in dataset_dict or "height" in dataset_dict:
        image_wh = (image.shape[1], image.shape[0])
        expected_wh = (dataset_dict["width"], dataset_dict["height"])
        if image_wh != expected_wh:
            raise SizeMismatchError(
                "Mismatched (W,H){}, got {}, expect {}".format(
                    " for image " + dataset_dict["file_name"]
                    if "file_name" in dataset_dict
                    else "",
                    image_wh,
                    expected_wh,
                )
            )

    # To ensure bbox always remap to original image size
    if "width" not in dataset_dict:
        dataset_dict["width"] = image.shape[1]
    if "height" not in dataset_dict:
        dataset_dict["height"] = image.shape[0]


def transform_proposals(dataset_dict, image_shape, transforms, *, proposal_topk, min_box_size=0):
    """
    Apply transformations to the proposals in dataset_dict, if any.

    Args:
        dataset_dict (dict): a dict read from the dataset, possibly
            contains fields "proposal_boxes", "proposal_objectness_logits", "proposal_bbox_mode"
        image_shape (tuple): height, width
        transforms (TransformList):
        proposal_topk (int): only keep top-K scoring proposals
        min_box_size (int): proposals with either side smaller than this
            threshold are removed

    The input dict is modified in-place, with abovementioned keys removed. A new
    key "proposals" will be added. Its value is an `Instances`
    object which contains the transformed proposals in its field
    "proposal_boxes" and "objectness_logits".
    """
    if "proposal_boxes" in dataset_dict:
        # Transform proposal boxes
        boxes = transforms.apply_box(
            BoxMode.convert(
                dataset_dict.pop("proposal_boxes"),
                dataset_dict.pop("proposal_bbox_mode"),
                BoxMode.XYXY_ABS,
            )
        )
        boxes = Boxes(boxes)
        objectness_logits = torch.as_tensor(
            dataset_dict.pop("proposal_objectness_logits").astype("float32")
        )

        boxes.clip(image_shape)
        keep = boxes.nonempty(threshold=min_box_size)
        boxes = boxes[keep]
        objectness_logits = objectness_logits[keep]

        proposals = Instances(image_shape)
        proposals.proposal_boxes = boxes[:proposal_topk]
        proposals.objectness_logits = objectness_logits[:proposal_topk]
        dataset_dict["proposals"] = proposals


def transform_instance_annotations(annotation, transforms, image_size, *, keypoint_hflip_indices=None):
    """
    Apply transforms to box, segmentation and keypoints annotations of a single instance.

    It will use `transforms.apply_box` for the box, and
    `transforms.apply_coords` for segmentation polygons & keypoints.
    If you need anything more specially designed for each data structure,
    you'll need to implement your own version of this function or the transforms.

    Args:
        annotation (dict): dict of instance annotations for a single instance.
            It will be modified in-place.
        transforms (TransformList):
        image_size (tuple): the height, width of the transformed image
        keypoint_hflip_indices (ndarray[int]): see `create_keypoint_hflip_indices`.

    Returns:
        dict:
            the same input dict with fields "bbox", "segmentation", "keypoints"
            transformed according to `transforms`.
            The "bbox_mode" field will be set to XYXY_ABS.
    """
    try_import_pycocotools()
    import pycocotools.mask as mask_util
    # bbox is 1d (per-instance bounding box)
    bbox = BoxMode.convert(annotation["bbox"], annotation["bbox_mode"], BoxMode.XYXY_ABS)
    # clip transformed bbox to image size
    bbox = transforms.apply_box([bbox])[0].clip(min=0)
    annotation["bbox"] = np.minimum(bbox, list(image_size + image_size)[::-1])
    annotation["bbox_mode"] = BoxMode.XYXY_ABS

    if "segmentation" in annotation:
        # each instance contains 1 or more polygons
        segm = annotation["segmentation"]
        if isinstance(segm, list):
            # polygons
            polygons = [np.asarray(p).reshape(-1, 2) for p in segm]
            annotation["segmentation"] = [
                p.reshape(-1) for p in transforms.apply_polygons(polygons)
            ]
        elif isinstance(segm, dict):
            # RLE
            mask = mask_util.decode(segm)
            mask = transforms.apply_segmentation(mask)
            assert tuple(mask.shape[:2]) == image_size
            annotation["segmentation"] = mask
        else:
            raise ValueError(
                "Cannot transform segmentation of type '{}'!"
                "Supported types are: polygons as list[list[float] or ndarray],"
                " COCO-style RLE as a dict.".format(type(segm))
            )

    if "keypoints" in annotation:
        keypoints = transform_keypoint_annotations(
            annotation["keypoints"], transforms, image_size, keypoint_hflip_indices
        )
        annotation["keypoints"] = keypoints

    if "beziers" in annotation:
        beziers = transform_beziers_annotations(annotation["beziers"], transforms)
        annotation["beziers"] = beziers

    return annotation


def transform_keypoint_annotations(keypoints, transforms, image_size, keypoint_hflip_indices=None):
    """
    Transform keypoint annotations of an image.
    If a keypoint is transformed out of image boundary, it will be marked "unlabeled" (visibility=0)

    Args:
        keypoints (list[float]): Nx3 float in Detectron2's Dataset format.
            Each point is represented by (x, y, visibility).
        transforms (TransformList):
        image_size (tuple): the height, width of the transformed image
        keypoint_hflip_indices (ndarray[int]): see `create_keypoint_hflip_indices`.
            When `transforms` includes horizontal flip, will use the index
            mapping to flip keypoints.
    """
    # (N*3,) -> (N, 3)
    keypoints = np.asarray(keypoints, dtype="float64").reshape(-1, 3)
    keypoints_xy = transforms.apply_coords(keypoints[:, :2])

    # Set all out-of-boundary points to "unlabeled"
    inside = (keypoints_xy >= np.array([0, 0])) & (keypoints_xy <= np.array(image_size[::-1]))
    inside = inside.all(axis=1)
    keypoints[:, :2] = keypoints_xy
    keypoints[:, 2][~inside] = 0

    # This assumes that HorizFlipTransform is the only one that does flip
    do_hflip = sum(isinstance(t, T.HFlipTransform) for t in transforms.transforms) % 2 == 1

    # Alternative way: check if probe points was horizontally flipped.
    # probe = np.asarray([[0.0, 0.0], [image_width, 0.0]])
    # probe_aug = transforms.apply_coords(probe.copy())
    # do_hflip = np.sign(probe[1][0] - probe[0][0]) != np.sign(probe_aug[1][0] - probe_aug[0][0])  # noqa

    # If flipped, swap each keypoint with its opposite-handed equivalent
    if do_hflip:
        assert keypoint_hflip_indices is not None
        keypoints = keypoints[keypoint_hflip_indices, :]

    # Maintain COCO convention that if visibility == 0 (unlabeled), then x, y = 0
    keypoints[keypoints[:, 2] == 0] = 0
    return keypoints


def annotations_to_instances(annos, image_size, mask_format="polygon"):
    """
    Create an :class:`Instances` object used by the models,
    from instance annotations in the dataset dict.

    Args:
        annos (list[dict]): a list of instance annotations in one image, each
            element for one instance.
        image_size (tuple): height, width

    Returns:
        Instances:
            It will contain fields "gt_boxes", "gt_classes",
            "gt_masks", "gt_keypoints", if they can be obtained from `annos`.
            This is the format that builtin models expect.
    """
    boxes = [BoxMode.convert(obj["bbox"], obj["bbox_mode"], BoxMode.XYXY_ABS) for obj in annos]
    target = Instances(image_size)
    target.gt_boxes = Boxes(boxes)

    classes = [obj["category_id"] for obj in annos]
    classes = torch.tensor(classes, dtype=torch.int64)
    target.gt_classes = classes

    if len(annos) and "segmentation" in annos[0]:
        segms = [obj["segmentation"] for obj in annos]
        if mask_format == "polygon":
            # TODO check type and provide better error
            masks = PolygonMasks(segms)
        else:
            try_import_pycocotools()
            import pycocotools.mask as mask_util
            assert mask_format == "bitmask", mask_format
            masks = []
            for segm in segms:
                if isinstance(segm, list):
                    # polygon
                    masks.append(polygons_to_bitmask(segm, *image_size))
                elif isinstance(segm, dict):
                    # COCO RLE
                    masks.append(mask_util.decode(segm))
                elif isinstance(segm, np.ndarray):
                    assert segm.ndim == 2, "Expect segmentation of 2 dimensions, got {}.".format(
                        segm.ndim
                    )
                    # mask array
                    masks.append(segm)
                else:
                    raise ValueError(
                        "Cannot convert segmentation of type '{}' to BitMasks!"
                        "Supported types are: polygons as list[list[float] or ndarray],"
                        " COCO-style RLE as a dict, or a full-image segmentation mask "
                        "as a 2D ndarray.".format(type(segm))
                    )
            # torch.from_numpy does not support array with negative stride.
            masks = BitMasks(
                torch.stack([torch.from_numpy(np.ascontiguousarray(x)) for x in masks])
            )
        target.gt_masks = masks

    if len(annos) and "keypoints" in annos[0]:
        kpts = [obj.get("keypoints", []) for obj in annos]
        target.gt_keypoints = Keypoints(kpts)

    # add attributes
    if len(annos) and "beziers" in annos[0]:
        beziers = [obj.get("beziers", []) for obj in annos]
        instance.beziers = torch.as_tensor(beziers, dtype=torch.float32)

    if len(annos) and "rec" in annos[0]:
        text = [obj.get("rec", []) for obj in annos]
        instance.text = torch.as_tensor(text, dtype=torch.int32)

    return target


# def annotations_to_instances_rotated(annos, image_size):
#     """
#     Create an :class:`Instances` object used by the models,
#     from instance annotations in the dataset dict.
#     Compared to `annotations_to_instances`, this function is for rotated boxes only

#     Args:
#         annos (list[dict]): a list of instance annotations in one image, each
#             element for one instance.
#         image_size (tuple): height, width

#     Returns:
#         Instances:
#             Containing fields "gt_boxes", "gt_classes",
#             if they can be obtained from `annos`.
#             This is the format that builtin models expect.
#     """
#     boxes = [obj["bbox"] for obj in annos]
#     target = Instances(image_size)
#     boxes = target.gt_boxes = RotatedBoxes(boxes)
#     boxes.clip(image_size)

#     classes = [obj["category_id"] for obj in annos]
#     classes = torch.tensor(classes, dtype=torch.int64)
#     target.gt_classes = classes

#     return target


def filter_empty_instances(instances, by_box=True, by_mask=True, box_threshold=1e-5):
    """
    Filter out empty instances in an `Instances` object.

    Args:
        instances (Instances):
        by_box (bool): whether to filter out instances with empty boxes
        by_mask (bool): whether to filter out instances with empty masks
        box_threshold (float): minimum width and height to be considered non-empty

    Returns:
        Instances: the filtered instances.
    """
    assert by_box or by_mask
    r = []
    if by_box:
        r.append(instances.gt_boxes.nonempty(threshold=box_threshold))
    if instances.has("gt_masks") and by_mask:
        r.append(instances.gt_masks.nonempty())

    # TODO: can also filter visible keypoints

    if not r:
        return instances
    m = r[0]
    for x in r[1:]:
        m = m & x
    return instances[m]


def create_keypoint_hflip_indices(dataset_names):
    """
    Args:
        dataset_names (list[str]): list of dataset names
    Returns:
        ndarray[int]: a vector of size=#keypoints, storing the
        horizontally-flipped keypoint indices.
    """

    check_metadata_consistency("keypoint_names", dataset_names)
    check_metadata_consistency("keypoint_flip_map", dataset_names)

    meta = MetadataCatalog.get(dataset_names[0])
    names = meta.keypoint_names
    # TODO flip -> hflip
    flip_map = dict(meta.keypoint_flip_map)
    flip_map.update({v: k for k, v in flip_map.items()})
    flipped_names = [i if i not in flip_map else flip_map[i] for i in names]
    flip_indices = [names.index(i) for i in flipped_names]
    return np.asarray(flip_indices)


def gen_crop_transform_with_instance(crop_size, image_size, instance):
    """
    Generate a CropTransform so that the cropping region contains
    the center of the given instance.

    Args:
        crop_size (tuple): h, w in pixels
        image_size (tuple): h, w
        instance (dict): an annotation dict of one instance, in Detectron2's
            dataset format.
    """
    crop_size = np.asarray(crop_size, dtype=np.int32)
    bbox = BoxMode.convert(instance["bbox"], instance["bbox_mode"], BoxMode.XYXY_ABS)
    center_yx = (bbox[1] + bbox[3]) * 0.5, (bbox[0] + bbox[2]) * 0.5
    assert (
        image_size[0] >= center_yx[0] and image_size[1] >= center_yx[1]
    ), "The annotation bounding box is outside of the image!"
    assert (
        image_size[0] >= crop_size[0] and image_size[1] >= crop_size[1]
    ), "Crop size is larger than image size!"

    min_yx = np.maximum(np.floor(center_yx).astype(np.int32) - crop_size, 0)
    max_yx = np.maximum(np.asarray(image_size, dtype=np.int32) - crop_size, 0)
    max_yx = np.minimum(max_yx, np.ceil(center_yx).astype(np.int32))

    y0 = np.random.randint(min_yx[0], max_yx[0] + 1)
    x0 = np.random.randint(min_yx[1], max_yx[1] + 1)
    return T.CropTransform(x0, y0, crop_size[1], crop_size[0])


def check_metadata_consistency(key, dataset_names):
    """
    Check that the datasets have consistent metadata.

    Args:
        key (str): a metadata key
        dataset_names (list[str]): a list of dataset names

    Raises:
        AttributeError: if the key does not exist in the metadata
        ValueError: if the given datasets do not have the same metadata values defined by key
    """
    if len(dataset_names) == 0:
        return
    logger = logging.getLogger(__name__)
    entries_per_dataset = [getattr(MetadataCatalog.get(d), key) for d in dataset_names]
    for idx, entry in enumerate(entries_per_dataset):
        if entry != entries_per_dataset[0]:
            logger.error(
                "Metadata '{}' for dataset '{}' is '{}'".format(key, dataset_names[idx], str(entry))
            )
            logger.error(
                "Metadata '{}' for dataset '{}' is '{}'".format(
                    key, dataset_names[0], str(entries_per_dataset[0])
                )
            )
            raise ValueError("Datasets have different metadata '{}'!".format(key))

def transform_beziers_annotations(beziers, transforms):
    """
    Transform keypoint annotations of an image.

    Args:
        beziers (list[float]): Nx16 float in Detectron2 Dataset format.
        transforms (TransformList):
    """
    # (N*2,) -> (N, 2)
    beziers = np.asarray(beziers, dtype="float64").reshape(-1, 2)
    beziers = transforms.apply_coords(beziers).reshape(-1)

    # This assumes that HorizFlipTransform is the only one that does flip
    do_hflip = (
        sum(isinstance(t, T.HFlipTransform) for t in transforms.transforms) % 2 == 1
    )
    if do_hflip:
        raise ValueError("Flipping text data is not supported (also disencouraged).")

    return beziers

def build_augmentation(cfg, is_train):
    """
    Create a list of default :class:`Augmentation` from config.
    Now it includes resizing and flipping.

    Returns:
        list[Augmentation]
    """
    if is_train:
        min_size = cfg.CONFIG.DATA.DETECTION.MIN_SIZE_TRAIN
        max_size = cfg.CONFIG.DATA.DETECTION.MAX_SIZE_TRAIN
        sample_style = cfg.CONFIG.DATA.DETECTION.MIN_SIZE_TRAIN_SAMPLING
    else:
        min_size = cfg.CONFIG.DATA.DETECTION.MIN_SIZE_TEST
        max_size = cfg.CONFIG.DATA.DETECTION.MAX_SIZE_TEST
        sample_style = "choice"
    if sample_style == "range":
        assert len(min_size) == 2, "more than 2 ({}) min_size(s) are provided for ranges".format(
            len(min_size)
        )

    augmentation = []
    augmentation.append(T.ResizeShortestEdge(min_size, max_size, sample_style))
    if is_train:
        augmentation.append(T.RandomFlip())
    return augmentation

def build_augmentation_ranged_clip(cfg, is_train):
    """
    With option to don't use hflip

    Returns:
        list[Augmentation]
    """
    if is_train:
        if cfg.CONFIG.DATA.DETECTION.MIN_SIZE_RANGE_TRAIN[0] == -1:
            min_size = cfg.CONFIG.DATA.DETECTION.MIN_SIZE_TRAIN
        else:
            assert len(cfg.CONFIG.DATA.DETECTION.MIN_SIZE_RANGE_TRAIN) == 2, \
                "MIN_SIZE_RANGE_TRAIN must have two elements (lower bound, upper bound)"
            min_size = range(
                cfg.CONFIG.DATA.DETECTION.MIN_SIZE_RANGE_TRAIN[0],
                cfg.CONFIG.DATA.DETECTION.MIN_SIZE_RANGE_TRAIN[1] + 1
            )
        max_size = cfg.CONFIG.DATA.DETECTION.MAX_SIZE_TRAIN
        sample_style = cfg.CONFIG.DATA.DETECTION.MIN_SIZE_TRAIN_SAMPLING
    else:
        min_size = cfg.CONFIG.DATA.DETECTION.MIN_SIZE_TEST
        max_size = cfg.CONFIG.DATA.DETECTION.MAX_SIZE_TEST
        sample_style = "choice"
    if sample_style == "range":
        assert (
            len(min_size) == 2
        ), "more than 2 ({}) min_size(s) are provided for ranges".format(len(min_size))

    logger = logging.getLogger(__name__)

    augmentation = []
    augmentation.append(T.ResizeShortestEdge(min_size, max_size, sample_style))
    if is_train:
        if cfg.CONFIG.DATA.DETECTION.HFLIP_TRAIN:
            augmentation.append(T.RandomFlip())
        logger.info("Augmentations used in training: " + str(augmentation))
    return augmentation
